from .order import Command as Order
from django_typer.management import group


class Command(Order):
    @group()
    def d(self):
        print("d")

    @d.command()
    def f(self):
        print("f")

    @d.command()
    def e(self):
        print("e")



from ..commands.upstream2 import Command as Upstream

# When plugging into an existing command we do not create
# a new class, but instead work directly with the commands
# and groups on the upstream class


# override init
@Upstream.initialize()
def init(self):
    return "plugin:init"


# override sub1
@Upstream.command()
def sub1(self):
    return "plugin:sub1"


# add a 3rd top level command
@Upstream.command()
def sub3(self):
    return "plugin:sub3"


# add a new subcommand to grp1
@Upstream.grp1.command()
def grp1_cmd2(self):
    return "plugin:grp1_cmd2"



from django_typer.management import Typer, TyperCommand
from django_typer.types import Verbosity

app = Typer()


@app.callback()
def init(self, verbosity: Verbosity = 0):
    assert isinstance(self, TyperCommand)
    self.verbosity = verbosity


@app.command()
def main(self, name: str):
    assert isinstance(self, TyperCommand)
    return {"verbosity": self.verbosity, "name": name}


grp2 = Typer()


@grp2.callback()
def init_grp1(self, flag: bool = False):
    assert isinstance(self, TyperCommand)
    self.flag = flag


@grp2.command()
def cmd2(self, fraction: float):
    assert isinstance(self, TyperCommand)
    return {"verbosity": self.verbosity, "flag": self.flag, "fraction": fraction}


app.add_typer(grp2, name="grp1")


@grp2.command()
def cmd1(self, count: int):
    assert isinstance(self, TyperCommand)
    return {"verbosity": self.verbosity, "flag": self.flag, "count": count}


def run_subgrp(self, msg: str):
    assert isinstance(self, TyperCommand)
    return {"verbosity": self.verbosity, "flag": self.flag, "msg": msg}


subgrp = Typer(name="subgrp", invoke_without_command=True, callback=run_subgrp)

grp2.add_typer(subgrp)



"""
Test options metavar propagation and overrides
"""

from django_typer.management import initialize
from .metavar2 import Command as Metavar2Command


class Command(Metavar2Command):
    @initialize()  # should inherit from Metavar2Command class metavar
    def init(self):
        print("init")



from django_typer.management import TyperCommand, command, group, initialize


class Command(TyperCommand):
    help = "Test staticmethods."

    @initialize(invoke_without_command=True)
    @staticmethod
    def init():
        return "init"

    @command()
    @staticmethod
    def cmd1():
        return "cmd1"

    @command()
    @staticmethod
    def cmd2():
        return "cmd2"

    @group(invoke_without_command=True)
    @staticmethod
    def grp1():
        return "grp1"

    @group(invoke_without_command=True)
    @staticmethod
    def grp2():
        return "grp2"

    @grp1.command()
    @staticmethod
    def grp1_cmd():
        return "grp1_cmd"

    @grp2.command()
    @staticmethod
    def grp2_cmd():
        return "grp2_cmd"

    @grp2.group(invoke_without_command=True)
    @staticmethod
    def grp2_subgrp():
        return "grp2_subgrp"

    @grp2_subgrp.command()
    @staticmethod
    def grp2_subgrp_cmd():
        return "grp2_subgrp_cmd"



from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestStatic(TestCase):
    cmd = "static"

    def test_run(self):
        stdout, stderr, retcode = run_command(self.cmd)
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "init")

        stdout, stderr, retcode = run_command(self.cmd, "cmd1")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "cmd1")

        stdout, stderr, retcode = run_command(self.cmd, "cmd2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "cmd2")

        stdout, stderr, retcode = run_command(self.cmd, "grp1")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp1")

        stdout, stderr, retcode = run_command(self.cmd, "grp2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp2")

        stdout, stderr, retcode = run_command(self.cmd, "grp1", "grp1-cmd")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp1_cmd")

        stdout, stderr, retcode = run_command(self.cmd, "grp2", "grp2-cmd")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp2_cmd")

        stdout, stderr, retcode = run_command(self.cmd, "grp2", "grp2-subgrp")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp2_subgrp")

        stdout, stderr, retcode = run_command(
            self.cmd, "grp2", "grp2-subgrp", "grp2-subgrp-cmd"
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "grp2_subgrp_cmd")

    def test_call(self):
        stdout = call_command(self.cmd)
        self.assertEqual(stdout.strip(), "init")

        stdout = call_command(self.cmd, "cmd1")
        self.assertEqual(stdout.strip(), "cmd1")

        stdout = call_command(self.cmd, "cmd2")
        self.assertEqual(stdout.strip(), "cmd2")

        stdout = call_command(self.cmd, "grp1")
        self.assertEqual(stdout.strip(), "grp1")

        stdout = call_command(self.cmd, "grp2")
        self.assertEqual(stdout.strip(), "grp2")

        stdout = call_command(self.cmd, "grp1", "grp1-cmd")
        self.assertEqual(stdout.strip(), "grp1_cmd")

        stdout = call_command(self.cmd, "grp2", "grp2-cmd")
        self.assertEqual(stdout.strip(), "grp2_cmd")

        stdout = call_command(self.cmd, "grp2", "grp2-subgrp")
        self.assertEqual(stdout.strip(), "grp2_subgrp")

        stdout = call_command(self.cmd, "grp2", "grp2-subgrp", "grp2-subgrp-cmd")
        self.assertEqual(stdout.strip(), "grp2_subgrp_cmd")

    def test_direct(self):
        from tests.apps.test_app.management.commands.static import (
            Command as Static,
        )

        static = get_command(self.cmd, Static)

        self.assertEqual(static.init(), "init")
        self.assertEqual(static.cmd1(), "cmd1")
        self.assertEqual(static.cmd2(), "cmd2")
        self.assertEqual(static.grp1(), "grp1")
        self.assertEqual(static.grp2(), "grp2")
        self.assertEqual(static.grp1_cmd(), "grp1_cmd")
        self.assertEqual(static.grp2_cmd(), "grp2_cmd")

        self.assertEqual(static.grp2_subgrp(), "grp2_subgrp")
        self.assertEqual(static.grp2_subgrp_cmd(), "grp2_subgrp_cmd")

        self.assertEqual(static.grp2.grp2_subgrp(), "grp2_subgrp")
        self.assertEqual(static.grp2.grp2_subgrp.grp2_subgrp_cmd(), "grp2_subgrp_cmd")


class TestStatic2(TestStatic):
    cmd = "static2"



from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestHandleAsInit(TestCase):
    def test_handle_as_init_run(self):
        stdout, stderr, retcode = run_command("handle_as_init")
        self.assertTrue("handle" in stdout)
        self.assertFalse(stderr.strip())
        self.assertEqual(retcode, 0)

        stdout, stderr, retcode = run_command("handle_as_init", "subcommand")
        self.assertTrue("subcommand" in stdout)
        self.assertFalse(stderr.strip())
        self.assertEqual(retcode, 0)

    def test_handle_as_init_call(self):
        self.assertEqual(call_command("handle_as_init").strip(), "handle")
        self.assertEqual(
            call_command("handle_as_init", "subcommand").strip(), "subcommand"
        )

    def test_handle_as_init_direct(self):
        self.assertEqual(get_command("handle_as_init")(), "handle")
        self.assertEqual(get_command("handle_as_init", "subcommand")(), "subcommand")
        self.assertEqual(get_command("handle_as_init").subcommand(), "subcommand")



from django_typer.management import Typer

app = Typer()


@app.command(name="subcommand1")
def handle():
    return "handle"


@app.command()
def subcommand2():
    return "subcommand2"


@app.command()
def subcommand3():
    return "subcommand3"



import typing as t

from django_typer.management import TyperCommand, command


class Command(TyperCommand, rich_markup_mode="markdown", chain=True):
    suppressed_base_arguments = [
        "--verbosity",
        "--traceback",
        "--no-color",
        "--force-color",
        "--skip_checks",
        "--settings",
        "--pythonpath",
        "--version",
    ]

    @command()
    def command1(self, option: t.Optional[str] = None):
        """This is a *markdown* help string"""
        assert self.__class__ is Command
        print("command1")
        return option

    @TyperCommand.command()
    def command2(self, option: t.Optional[str] = None):
        """This is a *markdown* help string"""
        assert self.__class__ is Command
        print("command2")
        return option



from django_typer.management import TyperCommand


class Command(TyperCommand):
    requires_system_checks = []
    requires_migrations_checks = False

    def handle(self, test_arg: int, print: bool = False):
        if print:
            import json

            return json.dumps({"app": "with_typer", "test_arg": test_arg})



from django_typer.management import TyperCommand


class Command(TyperCommand):
    help = "Return an object"

    def handle(self):
        assert self.__class__ is Command
        return {"key": "value"}



from django.utils.translation import gettext_lazy as _

from django_typer.management import TyperCommand


class Command(TyperCommand):
    help = "Test missing parameter."

    missing_args_message = _("{parameter} must be given.")

    def handle(self, arg1: int):
        pass



from tests.apps.test_app.management.commands.native_override_init import (
    Command,
    app,
)
from django_typer.types import Verbosity


@app.initialize()
def init(self, verbosity: Verbosity = 0, bog: bool = False):
    self.verbosity = verbosity
    self.bog = bog
    assert self.__class__ is Command
    return {"verbosity": self.verbosity, "bog": self.bog}


@app.group()
def grp2(self):
    """
    test_app2::grp2
    """
    assert self.__class__ is Command
    self.grp2_called = True


@app.grp2.command(name="cmd1")
def grp2_cmd1(self, g2arg1: int):
    """
    test_app2::grp2::grp2_cmd1
    """
    assert self.__class__ is Command
    return {
        "verbosity": self.verbosity,
        "bog": self.bog,
        "grp2_called": getattr(self, "grp2_called", None),
        "g2arg1": g2arg1,
    }



from django_typer.management import TyperCommand, command, finalize
from click import get_current_context


class Command(TyperCommand, chain=True):
    @finalize()
    def final(self, result):
        assert isinstance(self, Command)
        return "finalized: {}".format(result)

    @command()
    def cmd1(self, arg1: int = 1):
        return "cmd1 {}".format(arg1)

    @command()
    def cmd2(self, arg2: int):
        return "cmd2 {}".format(arg2)



import inspect
import os
import typing as t
from pathlib import Path

import typer
from django.conf import settings
from django.core.management import CommandError, call_command

from django_typer.management import (
    CommandNode,
    TyperCommand,
    command,
    initialize,
)
from django_typer.completers.db import databases as complete_db
from django_typer.completers.path import directories


class Command(TyperCommand):
    """
    Backup the website! This command groups backup routines together.
    Each routine may be run individually, but if no routine is specified,
    the default run of all routines will be executed.
    """

    suppressed_base_arguments = {"verbosity", "skip_checks"}

    requires_migrations_checks = False
    requires_system_checks = []

    databases = [alias for alias in settings.DATABASES.keys()]

    output_directory: Path

    @initialize(invoke_without_command=True)
    def init_or_run_all(
        self,
        # if we add a context argument Typer will provide it
        # the context is a click object that contains additional
        # information about the broader CLI invocation
        context: typer.Context,
        output_directory: t.Annotated[
            Path,
            typer.Option(
                "-o",
                "--output",
                shell_complete=directories,
                help="The directory to write backup artifacts to.",
            ),
        ] = Path(os.getcwd()),
    ):
        self.output_directory = output_directory

        if not self.output_directory.exists():
            self.output_directory.mkdir(parents=True)

        if not self.output_directory.is_dir():
            raise CommandError(f"{self.output_directory} is not a directory.")

        # here we use the context to determine if a subcommand was invoked and
        # if it was not we run all the backup routines
        if not context.invoked_subcommand:
            for cmd in self.get_backup_routines():
                cmd()

    def get_backup_routines(self) -> t.List[CommandNode]:
        """
        Return the list of backup subcommands. This is every registered command
        except for the list command.
        """
        # fetch all the command names at the top level of our command tree,
        # except for list, which we know to not be a backup routine
        return [
            cmd
            for name, cmd in self.get_subcommand().children.items()
            if name != "list"
        ]

    @command()
    def list(self):
        """
        List the default backup routines in the order they will be run.
        """
        self.echo("Default backup routines:")
        for cmd in self.get_backup_routines():
            sig = {
                name: param.default
                for name, param in inspect.signature(
                    cmd.callback
                ).parameters.items()
                if not name == "self"
            }
            params = ", ".join([f"{k}={v}" for k, v in sig.items()])
            self.secho(f"  {cmd.name}({params})", fg="green")

    @command()
    def database(
        self,
        filename: t.Annotated[
            str,
            typer.Option(
                "-f",
                "--filename",
                help=(
                    "The name of the file to use for the backup fixture. The "
                    "filename may optionally contain a {database} formatting "
                    "placeholder."
                ),
            ),
        ] = "{database}.json",
        databases: t.Annotated[
            t.List[str],
            typer.Option(
                "-d",
                "--database",
                help=(
                    "The name of the database(s) to backup. If not provided, "
                    "all databases will be backed up."
                ),
                shell_complete=complete_db(),
            ),
        ] = databases,
    ):
        """
        Backup database(s) to a json fixture file.
        """
        for db in databases or self.databases:
            output = self.output_directory / filename.format(database=db)
            self.echo(f"Backing up database [{db}] to: {output}")
            call_command(
                "dumpdata",
                output=output,
                database=db,
                format="json",
            )



import typing as t

from django.core.management.base import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import Typer
from tests.apps.examples.polls.models import Question as Poll


def get_poll_from_id(poll: t.Union[str, Poll]) -> Poll:
    if isinstance(poll, Poll):
        return poll
    try:
        return Poll.objects.get(pk=int(poll))
    except Poll.DoesNotExist:
        raise CommandError(f'Poll "{poll}" does not exist')


app = Typer(rich_markup_mode="markdown")


@app.command()
def handle(
    self,
    polls: t.Annotated[
        t.List[Poll],
        Argument(
            parser=get_poll_from_id,
            help=_("The database IDs of the poll(s) to close."),
        ),
    ],
    delete: t.Annotated[
        bool,
        Option(
            "--delete",  # we can also get rid of that unnecessary --no-delete flag
            help=_("Delete poll instead of closing it."),
        ),
    ] = False,
):
    """
    :sparkles: As mentioned in the last section, helps can also be set in
    the docstring and rendered using either
    [rich](https://rich.readthedocs.io/en/stable/markup.html)
    or [markdown](https://www.markdownguide.org/) :sparkles:
    """
    for poll in polls:
        poll.opened = False
        poll.save()
        self.stdout.write(
            self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
        )
        if delete:
            poll.delete()



import typing as t

from click import Context, Parameter
from click.core import ParameterSource
from click.shell_completion import CompletionItem
from django.apps import apps


def app_labels(
    ctx: Context, param: Parameter, incomplete: str
) -> t.List[CompletionItem]:
    """
    A case-sensitive completer for Django app labels or names. The completer
    prefers labels but names will also work.

    .. code-block:: python

        import typing as t
        import typer
        from django_typer.management import TyperCommand
        from django_typer.parsers import parse_app_label
        from django_typer.completers import complete_app_label

        class Command(TyperCommand):

            def handle(
                self,
                django_apps: t.Annotated[
                    t.List[AppConfig],
                    typer.Argument(
                        parser=parse_app_label,
                        shell_complete=complete_app_label,
                        help=_("One or more application labels.")
                    )
                ]
            ):
                ...

    :param ctx: The click context.
    :param param: The click parameter.
    :param incomplete: The incomplete string.
    :return: A list of matching app labels or names. Labels already present for the
        parameter on the command line will be filtered out.
    """
    present = []
    if (
        param.name
        and ctx.get_parameter_source(param.name) is not ParameterSource.DEFAULT
    ):
        present = [app.label for app in (ctx.params.get(param.name) or [])]
    ret = [
        CompletionItem(app.label)
        for app in apps.get_app_configs()
        if app.label.startswith(incomplete) and app.label not in present
    ]
    if not ret and incomplete:
        ret = [
            CompletionItem(app.name)
            for app in apps.get_app_configs()
            if app.name.startswith(incomplete)
            and app.name not in present
            and app.label not in present
        ]
    return ret



import json
import sys
import typing as t
from typing import Annotated


import typer
from django.utils.translation import gettext_lazy as _
from django.db.models import Q
from django.conf import settings
from django.utils.timezone import get_default_timezone

from django_typer.management import (
    TyperCommand,
    command,
    initialize,
)
from django_typer.utils import model_parser_completer
from django_typer import types
from django_typer.completers.model import text_query
from tests.apps.test_app.models import ShellCompleteTester


def obj_not_found(model_cls, value, exc):
    raise RuntimeError(f"Object {model_cls.__name__} with {value} not found.")


class Command(TyperCommand):
    @initialize()
    def init(self, verbosity: types.Verbosity = 1):
        assert self.__class__ is Command
        self.verbosity = verbosity

    @command()
    def test(
        self,
        char: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "char_field"),
                help=t.cast(str, _("Fetch objects by their char fields.")),
            ),
        ] = None,
        ichar: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester,
                    "char_field",
                    case_insensitive=True,
                    on_error=obj_not_found,
                ),
                help=t.cast(
                    str, _("Fetch objects by their char fields, case insensitive.")
                ),
            ),
        ] = None,
        text: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester,
                    "text_field",
                    query=text_query,
                ),
                help=t.cast(str, _("Fetch objects by their text fields.")),
            ),
        ] = None,
        itext: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester,
                    "text_field",
                    case_insensitive=True,
                    distinct=False,
                ),
                help=t.cast(
                    str, _("Fetch objects by their text fields, case insensitive.")
                ),
            ),
        ] = None,
        uuid: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "uuid_field"),
                help=t.cast(str, _("Fetch objects by their UUID fields.")),
            ),
        ] = None,
        id: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, order_by="id"),
                help=t.cast(str, _("Fetch objects by their int (pk) fields.")),
            ),
        ] = None,
        id_limit: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, limit=5),
                help=t.cast(str, _("Fetch objects by their int (pk) fields.")),
            ),
        ] = None,
        float: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "float_field"),
                help=t.cast(str, _("Fetch objects by their float fields.")),
            ),
        ] = None,
        decimal: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "decimal_field"),
                help=t.cast(str, _("Fetch objects by their decimal fields.")),
            ),
        ] = None,
        ip: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "ip_field"),
                help=t.cast(str, _("Fetch objects by their IP address fields.")),
            ),
        ] = None,
        email: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "email_field"),
                help=t.cast(str, _("Fetch objects by their email fields.")),
            ),
        ] = None,
        url: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "url_field"),
                help=t.cast(str, _("Fetch objects by their url fields.")),
            ),
        ] = None,
        filtered: Annotated[
            t.Optional[t.List[ShellCompleteTester]],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester.objects.filter(
                        ~(
                            Q(text_field__istartswith="a")
                            | Q(text_field__istartswith="s")
                        )
                    ),
                    "text_field",
                ),
                help=t.cast(str, _("Fetch objects by their text fields.")),
            ),
        ] = None,
        file: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "file_field"),
                help=t.cast(str, _("Fetch objects by their file fields.")),
            ),
        ] = None,
        file_path: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(ShellCompleteTester, "file_path_field"),
                help=t.cast(str, _("Fetch objects by their file path fields.")),
            ),
        ] = None,
        date: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester, "date_field", order_by=["-date_field"]
                ),
                help=t.cast(str, _("Fetch objects by their date fields.")),
            ),
        ] = None,
        datetime: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester,
                    "datetime_field",
                    order_by=("datetime_field", "id"),
                ),
                help=t.cast(str, _("Fetch objects by their datetime fields.")),
            ),
        ] = None,
        time: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester, "time_field", order_by="-time_field"
                ),
                help=t.cast(str, _("Fetch objects by their time fields.")),
            ),
        ] = None,
        duration: Annotated[
            t.Optional[ShellCompleteTester],
            typer.Option(
                **model_parser_completer(
                    ShellCompleteTester, "duration_field", order_by="duration_field"
                ),
                help=t.cast(str, _("Fetch objects by their duration fields.")),
            ),
        ] = None,
    ):
        assert self.__class__ is Command
        objects = {}
        if char is not None:
            assert isinstance(char, ShellCompleteTester)
            objects["char"] = {char.id: char.char_field}
        if ichar is not None:
            assert isinstance(ichar, ShellCompleteTester)
            objects["ichar"] = {ichar.id: ichar.char_field}
        if text:
            for txt in text:
                assert isinstance(txt, ShellCompleteTester)
            objects["text"] = [{txt.id: txt.text_field} for txt in text]
        if itext:
            for itxt in itext:
                assert isinstance(itxt, ShellCompleteTester)
            objects["itext"] = [{itxt.id: itxt.text_field} for itxt in itext]
        if uuid is not None:
            assert isinstance(uuid, ShellCompleteTester)
            objects["uuid"] = {uuid.id: str(uuid.uuid_field)}
        if id is not None:
            assert isinstance(id, ShellCompleteTester)
            objects["id"] = id.pk
        if float is not None:
            assert isinstance(float, ShellCompleteTester)
            objects["float"] = {float.id: str(float.float_field)}
        if decimal is not None:
            assert isinstance(decimal, ShellCompleteTester)
            objects["decimal"] = {decimal.id: str(decimal.decimal_field)}
        if ip is not None:
            for addr in ip:
                assert isinstance(addr, ShellCompleteTester)
            objects["ip"] = [{addr.id: addr.ip_field} for addr in ip]
        if filtered is not None:
            for txt in filtered:
                assert isinstance(txt, ShellCompleteTester)
            objects["filtered"] = [{txt.id: txt.text_field} for txt in filtered]
        if file is not None:
            assert isinstance(file, ShellCompleteTester)
            objects["file"] = {file.id: str(file.file_field)}
        if file_path is not None:
            assert isinstance(file_path, ShellCompleteTester)
            objects["file_path"] = {file_path.id: str(file_path.file_path_field)}
        if date is not None:
            assert isinstance(date, ShellCompleteTester)
            objects["date"] = {date.id: str(date.date_field)}
        if datetime is not None:
            assert isinstance(datetime, ShellCompleteTester)
            objects["datetime"] = {
                datetime.id: str(
                    datetime.datetime_field.astimezone(get_default_timezone())
                    if settings.USE_TZ
                    else datetime.datetime_field
                )
            }
        if time is not None:
            assert isinstance(time, ShellCompleteTester)
            objects["time"] = {time.id: str(time.time_field)}

        if duration is not None:
            assert isinstance(duration, ShellCompleteTester)
            objects["duration"] = {duration.id: str(duration.duration_field)}
        return json.dumps(objects)



import inspect
import json
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Tuple, Union
import re
import math
import time
from collections import Counter
import pexpect
from django.core.management.color import no_style
from django_typer import utils

# from charset_normalizer import from_bytes
import platform

rich_installed = utils.rich_installed


TESTS_DIR = Path(__file__).parent
DJANGO_PARAMETER_LOG_FILE = TESTS_DIR / "dj_params.json"
manage_py = TESTS_DIR.parent / "manage.py"
WORD = re.compile(r"\w+")


def get_cosine(vec1, vec2):
    intersection = set(vec1.keys()) & set(vec2.keys())
    numerator = sum([vec1[x] * vec2[x] for x in intersection])

    sum1 = sum([vec1[x] ** 2 for x in list(vec1.keys())])
    sum2 = sum([vec2[x] ** 2 for x in list(vec2.keys())])
    denominator = math.sqrt(sum1) * math.sqrt(sum2)

    if not denominator:
        return 0.0
    else:
        return float(numerator) / denominator


def text_to_vector(text):
    words = WORD.findall(text)
    return Counter(words)


def similarity(text1, text2):
    """
    Compute the cosine similarity between two texts.
    https://en.wikipedia.org/wiki/Cosine_similarity

    We use this to lazily evaluate the output of --help to our
    renderings.
    #"""
    vector1 = text_to_vector(text1)
    vector2 = text_to_vector(text2)
    return get_cosine(vector1, vector2)


def strip_control_sequences(text):
    """
    Removes control sequences, such as ANSI escape codes, from the given text.

    Args:
        text (str): The input text containing control sequences.

    Returns:
        str: The cleaned text without control sequences.
    """
    # Regular expression to match ANSI escape codes
    ansi_escape_pattern = re.compile(r"\x1b(?:[@-_][0-?]*[ -/]*[@-~])")
    return ansi_escape_pattern.sub("", text)


def get_named_arguments(function):
    sig = inspect.signature(function)
    return [
        name
        for name, param in sig.parameters.items()
        if param.default != inspect.Parameter.empty
    ]


def get_named_defaults(function):
    sig = inspect.signature(function)
    return {
        name: param.default
        for name, param in sig.parameters.items()
        if param.default != inspect.Parameter.empty
    }


if platform.system() == "Windows":
    import winpty

    class WinPtyWrapper:
        strip_ansi = True
        buffer = ""

        def __init__(self, command, env=None, cwd=None, strip_ansi=strip_ansi):
            self.strip_ansi = strip_ansi
            self.buffer = ""
            self.process = winpty.PtyProcess.spawn(command, cwd=cwd, env=env)

        def sendline(self, line):
            self.process.write(line + "\r\n")

        def read(self):
            return self.process.read().encode("utf-8")

        def expect(self, pattern, timeout=None):
            if not self.isalive():
                raise EOFError("Process terminated before finding the pattern.")

            start_time = time.time()

            while True:
                if timeout is not None and (time.time() - start_time) > timeout:
                    raise TimeoutError(f"Pattern '{pattern}' not found within timeout.")

                chunk = self.read().decode("utf-8")
                if self.strip_ansi:
                    chunk = strip_control_sequences(chunk)
                self.buffer += chunk
                if self.buffer:
                    # Search for the pattern in the buffer
                    match = re.search(pattern, self.buffer)
                    if match:
                        self.buffer = self.buffer[match.end() :]
                        return True

                # Check if the process has ended
                if not self.isalive():
                    raise EOFError("Process terminated before finding the pattern.")

        def isalive(self):
            return self.process.isalive()

        def close(self):
            self.process.close()


def interact(command, *args, **kwargs):
    if platform.system() == "Windows":
        return WinPtyWrapper(
            " ".join([sys.executable, f"./{manage_py.name}", command, *args]),
            env=os.environ,
            cwd=str(manage_py.parent),
        )
    else:
        cwd = os.getcwd()
        try:
            os.chdir(manage_py.parent)
            return pexpect.spawn(
                " ".join([sys.executable, f"./{manage_py.name}", command, *args]),
                env=os.environ,
                **kwargs,
            )
        finally:
            os.chdir(cwd)


def run_command(
    command, *args, parse_json=True, chdir=True, time=False, **kwargs
) -> Union[Tuple[str, str, int], Tuple[str, str, int, float]]:
    # we want to use the same test database that was created for the test suite run
    cwd = os.getcwd()
    try:
        env = kwargs.pop("env") if "env" in kwargs else os.environ.copy()
        if platform.system() == "Windows":
            env["PYTHONIOENCODING"] = "utf-8"
            env["PYTHONUTF8"] = "1"
        if chdir:
            os.chdir(manage_py.parent)
        cmd = [
            sys.executable,
            f"./{manage_py.name}" if chdir else manage_py,
            command,
            *args,
        ]
        time_seconds = 0.0
        if time:
            cmd.insert(0, "time")
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=False,
            env=env,
            **kwargs,
        )
        stdout_encoding = (
            "utf-8"  # getattr(from_bytes(result.stdout).best(), "encoding", "utf-8")
        )
        stdout = result.stdout.decode(stdout_encoding)
        stderr_encoding = (
            "utf-8"  # getattr(from_bytes(result.stderr).best(), "encoding", "utf-8")
        )
        stderr = result.stderr.decode(stderr_encoding)
        if time:
            time_seconds = float(stderr.split("real")[0].strip())

        # Check the return code to ensure the script ran successfully
        if result.returncode != 0:
            return stdout, stderr, result.returncode

        # Parse the output
        if result.stdout:
            if parse_json:
                try:
                    if time:
                        return (
                            json.loads(stdout),
                            stderr,
                            result.returncode,
                            time_seconds,
                        )
                    else:
                        return json.loads(stdout), stderr, result.returncode
                except json.JSONDecodeError:
                    if time:
                        return stdout, stderr, result.returncode, time_seconds
                    return stdout, stderr, result.returncode
            if time:
                return stdout, stderr, result.returncode, time_seconds
            return stdout, stderr, result.returncode
        if time:
            return stdout, stderr, result.returncode, time_seconds
        return stdout, stderr, result.returncode
    finally:
        os.chdir(cwd)


def log_django_parameters(django_command, **extra):
    if DJANGO_PARAMETER_LOG_FILE.exists():
        DJANGO_PARAMETER_LOG_FILE.unlink()

    from django.conf import settings

    with open(DJANGO_PARAMETER_LOG_FILE, "w") as f:
        json.dump(
            {
                "settings": settings.SETTINGS_FILE,
                "python_path": sys.path,
                "no_color": django_command.style == no_style(),
                "no_color_attr": django_command.no_color,
                "force_color_attr": django_command.force_color,
                **extra,
            },
            f,
            indent=4,
        )


def read_django_parameters():
    try:
        with open(DJANGO_PARAMETER_LOG_FILE) as f:
            return json.load(f)
    except FileNotFoundError:
        return {}
    finally:
        if DJANGO_PARAMETER_LOG_FILE.exists():
            DJANGO_PARAMETER_LOG_FILE.unlink()


def to_platform_str(path: str) -> str:
    return path.replace("/", os.path.sep)



"""
Django Typer app config. This module includes settings check and rich traceback
installation logic.
"""

import inspect
import typing as t

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import CheckMessage, register
from django.core.checks import Warning as CheckWarning

from django_typer import patch

from .config import traceback_config, use_rich_tracebacks
from .utils import install_traceback

patch.apply()
install_traceback()


@register("settings")
def check_traceback_config(app_configs, **kwargs) -> t.List[CheckMessage]:
    """
    A system check that validates that the traceback config is valid and
    contains only the expected parameters.
    """

    warnings: t.List[CheckMessage] = []
    if use_rich_tracebacks():
        tb_config = traceback_config()
        from rich import traceback  # pyright: ignore[reportMissingImports]

        expected = {
            "no_install",
            "short",
            *inspect.signature(traceback.install).parameters.keys(),
        }
        unexpected = set(tb_config.keys()) - expected
        if unexpected:
            warnings.append(
                CheckWarning(
                    "DT_RICH_TRACEBACK_CONFIG",
                    hint="Unexpected parameters encountered: {keys}.".format(
                        keys=", ".join(unexpected)
                    ),
                    obj=settings.SETTINGS_MODULE,
                    id="django_typer.W001",
                )
            )
    return warnings


class DjangoTyperConfig(AppConfig):
    """
    Django Typer app config.
    """

    name = "django_typer"
    label = name
    verbose_name = "Django Typer"

    def ready(self):
        from django_typer.management import extensions
        from django_typer.utils import register_command_plugins

        register_command_plugins(extensions, ["shellcompletion"])



from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand):
    help = "Takes no arguments"

    suppressed_base_arguments = [
        "verbosity",
        "no_color",
        "force_color",
        "help",
        "settings",
        "pythonpath",
        "traceback",
        "version",
        "skip_checks",
    ]

    @initialize()
    def init(self):
        assert self.__class__ is Command

    @command(add_help_option=False)
    def cmd(self):
        assert self.__class__ is Command



from django_typer.management import Typer

# alternatively we can use the finalize nomenclature of the TyperCommand
# interface - this is a non-standard Typer extension

app = Typer(chain=True)


# The Typer interface is extended with the finalize decorator
@app.finalize()
def to_csv(results):
    return ", ".join(results)


@app.command()
def cmd1():
    return "result1"


@app.command()
def cmd2():
    return "result2"



from django_typer.management import command, group, initialize
from tests.apps.test_app.management.commands.interference import (
    Command as Interference,
)


class Command(Interference):
    help = "Inherited and extend interference"

    @initialize(invoke_without_command=True)
    def init(self, arg: bool = True):
        assert self.__class__ is Command
        return f"test_app2::interference::init({arg})"

    @command()
    def cmd1(self, arg1: int):
        """
        Command 1, takes an int.
        """
        assert self.__class__ is Command
        return f"test_app2::interference::cmd1({arg1})"

    @command()
    def cmd_ext(self):
        """
        Command EXT.
        """
        assert self.__class__ is Command
        return "test_app2::interference::cmd_ext()"

    @Interference.command()
    def adapt_cmd(self):
        """Should be equivalent to @command()"""
        assert self.__class__ is Command
        return "test_app2::interference::adapt_cmd()"

    @Interference.group(invoke_without_command=True)
    def adapt_group(self):
        """Should be equivalent to @group()"""
        assert self.__class__ is Command
        return "test_app2::interference::adapt_group()"

    @group(invoke_without_command=True)
    def grp2(self, argg2: str):
        """
        Group 2, take a str.
        """
        assert self.__class__ is Command
        return f"test_app2::interference::grp2({argg2})"

    @Interference.grp1.command()
    def cmd3(self, arg3_0: int, arg3_1: int):
        """
        Command 3, take two ints.
        """
        assert self.__class__ is Command
        return f"test_app2::interference::grp1({self.argg3})::cmd3({arg3_0}, {arg3_1})"

    @Interference.subgroup.command()
    def cmd5(self):
        assert self.__class__ is Command
        return f"test_app2::interference::grp1({self.argg3})::subgroup({self.arg5_0})::cmd5()"

    @Interference.subsubgroup.command()
    def subsubcommand(self):
        assert self.__class__ is Command
        return f"test_app2::interference::grp1({self.argg3})::subgroup({self.arg5_0})::subsubgroup({self.subsubgroup_called})::subsubcommand()"

    @Interference.subsubgroup.command()
    def subsubcommand2(self):
        assert self.__class__ is Command
        return f"test_app2::interference::grp1({self.argg3})::subgroup({self.arg5_0})::subsubgroup({self.subsubgroup_called})::subsubcommand2()"



import shutil
from pathlib import Path
import typing as t

import pytest
from django.test import TestCase, override_settings
import sys
import os
import platform

from tests.shellcompletion import (
    _ScriptCompleteTestCase,
    _InstalledScriptCompleteTestCase,
)


@pytest.mark.skipif(shutil.which("zsh") is None, reason="Z-Shell not available")
class ZshTests(_ScriptCompleteTestCase, TestCase):
    shell = "zsh"
    directory = Path("~/.zfunc").expanduser()
    interactive_opt = "-i"
    tabs = "\t\t\t"

    environment = [
        f"PATH={Path(sys.executable).parent}:$PATH",
        f"DJANGO_SETTINGS_MODULE=tests.settings.completion",
    ]

    def verify_install(self, script=None, directory: t.Optional[Path] = None):
        directory = directory or self.directory
        if not script:
            script = self.manage_script
        self.assertTrue((directory / f"_{script}").exists())

    def verify_remove(self, script=None, directory: t.Optional[Path] = None):
        directory = directory or self.directory
        if not script:
            script = self.manage_script
        self.assertFalse((directory / f"_{script}").exists())


@pytest.mark.skipif(shutil.which("zsh") is None, reason="Z-Shell not available")
class ZshExeTests(_InstalledScriptCompleteTestCase, ZshTests, TestCase):
    shell = "zsh"

    @pytest.mark.skipif(
        not bool(os.environ.get("ENABLE_CI_ONLY_TESTS", False)),
        reason="This test is dangerous to run on a user machine, "
        "because it may nuke their shell profile file.",
    )
    def test_no_zshrc_file(self):
        zshrc = ""
        try:
            if (Path.home() / ".zshrc").exists():
                zshrc = (Path.home() / ".zshrc").read_text()
                os.unlink(Path.home() / ".zshrc")
            self.test_shell_complete()
            os.unlink(Path.home() / ".zshrc")
            self.remove()
            self.verify_remove()
        finally:
            if zshrc:
                (Path.home() / ".zshrc").write_text(zshrc)

    @override_settings(TEMPLATES=[])
    def test_no_template_config(self):
        self.test_shell_complete()

    if platform.system() != "Windows":

        def test_prompt_install(self, env={}, directory=None):
            zdot_dir = Path(__file__).parent / "zdotdir"
            try:
                zdot_dir.mkdir(exist_ok=True)
                super().test_prompt_install(
                    env={"ZDOTDIR": str(zdot_dir.absolute())},
                    directory=zdot_dir / ".zfunc",
                )
            finally:
                if zdot_dir.exists():
                    shutil.rmtree(zdot_dir)



from django_typer.management import Typer
from django_typer.types import Verbosity

app = Typer()

# after the first Typer() call this module will have a Command class
# and we can modify it directly to remove the --settings option
assert app.django_command
app.django_command.suppressed_base_arguments = ["--settings"]


@app.command()
def handle(verbosity: Verbosity = 1): ...



import subprocess
import typing as t
from pathlib import Path

from click.shell_completion import CompletionItem

from . import DjangoTyperShellCompleter


class PowerShellComplete(DjangoTyperShellCompleter):
    """
    This completer class supports the PowerShell_ versions < 6.
    See :class:`~PwshComplete` for versions >= 6.

    Completion scripts are installed into the user's profile file.

    Returned suggestions are delimited by ``:::``. Each suggestion is one one line.
    """

    name = "powershell"
    """
    shell executable.
    """

    template = "shell_complete/powershell.ps1"
    """
    The template used to render the powershell completion script.
    """

    supports_scripts = False
    """
    PowerShell does not support script invocations.
    """

    color = False
    """
    PowerShell_ does support ansi control codes in completion suggestions, but we
    disable them by default.
    """

    def format_completion(self, item: CompletionItem) -> str:
        return ":::".join(
            [
                item.type,
                self.process_rich_text(item.value),
                " ".join(
                    [
                        ln.strip()
                        for ln in self.process_rich_text(item.help or " ").splitlines()
                    ]
                ),
            ]
        )

    def set_execution_policy(self) -> None:
        subprocess.run(
            [
                self.name,
                "-Command",
                "Set-ExecutionPolicy",
                "Unrestricted",
                "-Scope",
                "CurrentUser",
            ]
        )

    def get_user_profile(self) -> Path:
        result = subprocess.run(
            [self.name, "-NoProfile", "-Command", "echo", "$profile"],
            check=True,
            stdout=subprocess.PIPE,
        )
        if result.returncode == 0:
            for encoding in ["windows-1252", "utf8", "cp850"]:
                try:
                    return Path(result.stdout.decode(encoding).strip())
                except UnicodeDecodeError:  # pragma: no cover
                    pass
        raise RuntimeError(
            "Unable to find the PowerShell user profile."
        )  # pragma: no cover

    def install(self, prompt: bool = True) -> t.List[Path]:
        assert self.prog_name
        self.uninstall()
        self.set_execution_policy()
        profile = self.get_user_profile()
        start_line = 0 if not profile.exists() else profile.read_text().count("\n") + 1
        source = self.source()
        if self.prompt(
            prompt=prompt, source=source, file=profile, start_line=start_line
        ):
            profile.parent.mkdir(parents=True, exist_ok=True)
            with profile.open(mode="a") as f:
                f.writelines(["", self.source()])
            return [profile]
        return []

    def uninstall(self) -> None:
        # annoyingly, powershell has one profile script for all completion commands
        # so we have to find our entry and remove it
        assert self.prog_name
        self.set_execution_policy()
        profile = self.get_user_profile()
        if not profile.exists():
            return
        edited_lines = []
        mark = None
        with open(profile, "rt", encoding="utf-8") as pwr_sh:
            for line in pwr_sh.readlines():
                edited_lines.append(line)
                if line.startswith("Import-Module PSReadLine"):
                    mark = len(edited_lines) - 1
                elif (
                    mark is not None
                    and line.startswith("Register-ArgumentCompleter")
                    and f" {self.command.manage_script_name} " in line
                ):
                    edited_lines = edited_lines[:mark]
                    mark = None

        if edited_lines:
            with open(profile, "wt", encoding="utf-8") as pwr_sh:
                pwr_sh.writelines(edited_lines)
        else:
            profile.unlink()


class PwshComplete(PowerShellComplete):
    """
    This completer class supports the PowerShell_ versions >= 6. See
    :class:`~PowerShellComplete` for versions < 6.

    All behaviors are the same as :class:`~PowerShellComplete`, except that ``pwsh`` is
    used instead of ``powershell``.
    """

    name = "pwsh"
    """
    shell executable.
    """



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from typer import Argument, Option
import typing as t

from django_typer.management import command, group, initialize
from django_typer import types
from tests.apps.test_app.management.commands.groups import (
    Command as GroupsCommand,
)


class Command(GroupsCommand, epilog="Overridden from test_app."):
    help = "Test groups command inheritance."

    precision = 2
    verbosity = 1

    suppressed_base_arguments = ["--version"]

    setting_name: str

    @initialize()
    def init(self, verbosity: types.Verbosity = verbosity):
        """
        Initialize the command.
        """
        assert self.__class__ is Command
        self.verbosity = verbosity

    @command()
    def echo(
        self,
        message: str,
        echoes: t.Annotated[
            int, Argument(help="Number of times to echo the message.")
        ] = 1,
    ):
        """
        Echo the given message the given number of times.
        """
        assert self.__class__ is Command
        return " ".join([message] * echoes)

    # test override base class command and remove arguments
    @GroupsCommand.string.case.command()
    def upper(self) -> str:
        assert self.__class__ is Command
        return super().upper(0, None)

    @GroupsCommand.string.command()
    def strip(self):
        """Strip white space off the ends of the string"""
        assert self.__class__ is Command
        return self.op_string.strip()

    @group()
    def setting(
        self, setting: t.Annotated[str, Argument(help=_("The setting variable name."))]
    ):
        """
        Get or set Django settings.
        """
        assert self.__class__ is Command
        self.setting_name = setting

    @setting.command()
    def print(
        self,
        safe: t.Annotated[bool, Option(help=_("Do not assume the setting exists."))],
    ):
        """
        Print the setting value.
        """
        assert self.__class__ is Command
        if safe:
            return getattr(settings, self.setting_name, None)
        return getattr(settings, self.setting_name)



from django.core.management import call_command
from django.test import TestCase
import pytest

from tests.utils import interact


class TestPromptOptions(TestCase):
    @pytest.mark.timeout(30)
    def test_run_with_option_prompt(self):
        cmd = interact("prompt", "--no-color", "cmd1", "bckohan")
        cmd.expect("Password:")
        cmd.sendline("test_password")
        cmd.expect("bckohan test_password")

        cmd = interact("prompt", "--no-color", "cmd2", "bckohan")
        cmd.expect("bckohan None")

        cmd = interact("prompt", "--no-color", "cmd2", "bckohan", "-p")
        cmd.expect("Password:")
        cmd.sendline("test_password2")
        cmd.expect("bckohan test_password2")

        cmd = interact("prompt", "--no-color", "cmd3", "bckohan")
        cmd.expect("bckohan default")

        cmd = interact("prompt", "--no-color", "cmd3", "bckohan", "-p")
        cmd.expect(r"Password \[default\]:")
        cmd.sendline("test_password3")
        cmd.expect("bckohan test_password3")

        cmd = interact("prompt", "--no-color", "group1", "cmd4", "bckohan")
        cmd.expect(r"Flag:")
        cmd.sendline("test_flag")
        cmd.expect(r"Password:")
        cmd.sendline("test_password4")
        cmd.expect("test_flag bckohan test_password4")

    def test_call_with_option_prompt(self):
        self.assertEqual(
            call_command(
                "prompt", "--no-color", "cmd1", "bckohan", password="test_password"
            ),
            "bckohan test_password",
        )

        self.assertEqual(
            call_command("prompt", "--no-color", "cmd2", "bckohan"), "bckohan None"
        )

        self.assertEqual(
            call_command(
                "prompt", "--no-color", "cmd2", "bckohan", "-p", "test_password2"
            ),
            "bckohan test_password2",
        )

        self.assertEqual(
            call_command("prompt", "--no-color", "cmd3", "bckohan"), "bckohan default"
        )

        self.assertEqual(
            call_command(
                "prompt", "--no-color", "cmd3", "bckohan", password="test_password3"
            ),
            "bckohan test_password3",
        )

        self.assertEqual(
            call_command(
                "prompt",
                "--no-color",
                "group1",
                "-f",
                "test_flag",
                "cmd4",
                "bckohan",
                password="test_password4",
            ),
            "test_flag bckohan test_password4",
        )

    def test_call_group_with_prompt_value(self):
        """
        This is a bug!
        """
        self.assertEqual(
            call_command(
                "prompt",
                "--no-color",
                "group1",
                "cmd4",
                "bckohan",
                flag="test_flag",
                password="test_password4",
            ),
            "test_flag bckohan test_password4",
        )



from functools import partial

from django.conf import settings

from . import these_strings

# use a function that returns a generator because we should not access settings on
# import
databases = partial(these_strings, lambda: settings.DATABASES.keys())
"""
A completer that completes Django database aliases configured in settings.DATABASES.

:param allow_duplicates: Whether or not to allow duplicate values. Defaults to False.
:return: A completer function.
"""



#!python

import bisect
import json
import os
import platform
import shutil
import subprocess
import sys
import typing as t
from dataclasses import asdict, dataclass, is_dataclass
from hashlib import sha256
from pathlib import Path
from pprint import pprint

import django
import typer

from django_typer import VERSION as django_typer_version

app = typer.Typer()


def dataclass_encoder(obj):
    if is_dataclass(obj):
        return asdict(obj)
    raise TypeError(f"Type {type(obj)} not serializable")


@dataclass(frozen=True)
class EnvKey:
    django_typer: t.Tuple[int, int, int]
    python: t.Tuple[int, int]
    django: t.Tuple[int, int]
    typer: t.Tuple[int, int]

    def astuple(self):
        return (self.django_typer, self.python, self.django, self.typer)


@dataclass(frozen=True)
class Platform:
    os: str
    arch: str
    name: str


@dataclass(frozen=True)
class RunKey:
    cmd: str
    """
    The command that was run.
    """

    typer: bool
    """
    Is this a Typer command?
    """

    app: bool
    """
    Is the django_typer app installed?
    """

    rich: bool
    """
    Is rich installed?
    """

    help: bool
    """
    Is this a help run?
    """

    def __hash__(self):
        """
        Python randomizes hashes across interpreter runs so we need to define our own
        repeatable hash
        """
        return int.from_bytes(
            sha256(
                f"{self.cmd}|{int(self.typer)}|{int(self.app)}|"
                f"{int(self.rich)}|{int(self.help)}".encode("utf-8")
            ).digest(),
            byteorder="big",
        )


django_admin = os.path.join(os.environ["VIRTUAL_ENV"], "bin", "django-admin")

no_typer = {"DJANGO_SETTINGS_MODULE": "profiling.settings.no_typer"}
with_typer = {"DJANGO_SETTINGS_MODULE": "profiling.settings.with_typer"}
with_typer_app = {"DJANGO_SETTINGS_MODULE": "profiling.settings.with_typer_app"}
polls = {"DJANGO_SETTINGS_MODULE": "profiling.settings.polls"}
polls_no_typer = {"DJANGO_SETTINGS_MODULE": "profiling.settings.polls_no_typer"}
polls_with_typer = {"DJANGO_SETTINGS_MODULE": "profiling.settings.polls_with_typer"}

pythonpath = os.path.realpath(__file__ + "/../..")

run_counts = 50

rich_installed = False
try:
    import rich  # noqa: F401

    rich_installed = True
except ImportError:
    pass

profiling_dir = os.path.realpath(__file__ + "/..")
profile_file = os.path.realpath(__file__ + "/../profile.json")


def get_platform() -> Platform:
    if shutil.which("scutil"):
        computer = (
            subprocess.check_output(["scutil", "--get", "ComputerName"])
            .strip()
            .decode()
        )
    else:
        computer = platform.node()
    return Platform(
        os=platform.system().lower(), arch=platform.machine().lower(), name=computer
    )


def run_command(*cmd, **env):
    result = subprocess.run(
        [*cmd],
        capture_output=True,
        text=True,
        env={"PYTHONPATH": pythonpath, "VIRTUAL_ENV": os.environ["VIRTUAL_ENV"], **env},
    )
    stdout = result.stdout
    if stdout:
        import json

        try:
            stdout = json.loads(result.stdout)
        except json.JSONDecodeError:
            pass
    return stdout, result.stderr, result.returncode


def time_command(*cmd, **env):
    stdout, stderr, retcode = run_command("time", *cmd, **env)
    real_seconds = float(stderr.split("real")[0]) if stderr else None
    return stdout, stderr, retcode, real_seconds


def import_trace(*cmd, save=None, **env):
    _, stderr, _ = run_command(sys.executable, "-X", "importtime", *cmd, **env)
    imports = {}
    for line in stderr.splitlines()[1:]:
        line = line.strip()
        if line and line.startswith("import time:"):
            slf, cum, pkg = line[len("import time: ") :].strip().split("|")
            slf = int(slf.strip())
            cum = int(cum.strip())
            pkg = pkg.strip()
            imports[pkg] = (slf, cum)
    if save:
        with open(profiling_dir + "/" + save.replace("_.", "."), "wt") as f:
            f.write(stderr)
    return imports


def load():
    if os.path.exists(profile_file):
        with open(profile_file, "r") as f:
            profile_data = json.load(f)
            for key, value in profile_data.items():
                value["env"] = EnvKey(**{k: tuple(v) for k, v in value["env"].items()})
                value["platform"] = Platform(**value["platform"])
                profile_data[key] = value
                for id, run in value["runs"].items():
                    run["run"] = RunKey(**run["run"])
                    assert id == str(hash(run["run"]))
            return profile_data
    return {}


@app.command(help="Load fixture data.")
def load_fixtures():
    stdout, stderr, retcode = run_command(django_admin, "migrate", **polls)
    if retcode == 0:
        print(stdout)
    else:
        print(stderr)
    stdout, stderr, retcode = run_command(
        django_admin, "loaddata", str(Path(__file__).parent / "polls.json"), **polls
    )
    if retcode == 0:
        print(stdout)
    else:
        print(stderr)


@app.command(
    help="Generate profiling data for django_typer in the current environment."
)
def generate():
    load_fixtures()
    profile_data = load()

    env_key = EnvKey(
        django_typer=django_typer_version,
        python=sys.version_info[0:2],
        django=django.VERSION[:2],
        typer=tuple(int(v) for v in typer.__version__.split(".")[:2]),
    )

    # json files dont allow tuples to be keys so to keep the latest data at the end of
    # the list we have to do a little manual sorting
    env_keys = sorted([run["env"].astuple() for run in profile_data.values()])
    pos = bisect.bisect_left(env_keys, env_key.astuple()) or 0
    pos_str = str(pos)
    if pos_str in profile_data:
        if profile_data[pos_str]["env"] != env_key:
            for idx in reversed(
                range(pos, max((int(i) for i in profile_data.keys())) + 1)
            ):
                if str(idx) in profile_data:
                    profile_data[str(idx + 1)] = profile_data.pop(str(idx))

    profile_data.setdefault(pos_str, {"env": env_key, "runs": {}})
    profile_data[pos_str]["platform"] = get_platform()

    last_idx = pos - 1
    while last_idx >= 0 and str(last_idx) not in profile_data:
        last_idx -= 1

    for hlp in [False, True]:
        for cmd, env in [
            (("minimal", "5"), no_typer),
            (("minimal", "5"), with_typer),
            (("minimal", "5"), with_typer_app),
            (("minimal", "5"), with_typer_app),
            (("shellcompletion", "complete", "polls "), polls_with_typer),
            (("polls", "1"), polls_no_typer),
            (("polls", "2"), polls_with_typer),
        ]:
            run = RunKey(
                cmd=cmd[0],
                typer=not (
                    env is no_typer or env is polls_no_typer
                ),  # is this a typer command?
                app=env is with_typer_app
                or env is polls_with_typer,  # is django_typer in INSTALLED_APPS?
                rich=rich_installed,  # is rich installed?
                help=hlp,  # is this a help run?
            )
            if not run.typer and run.rich:
                continue  # no need for this combo

            # skip shellcompletion help
            if hlp and cmd[0] == "shellcompletion":
                continue

            if cmd[0] == "minimal":
                # sanity checks
                stats: t.Dict[str, t.Any] = run_command(
                    django_admin, "minimal", "5", "--print", **env
                )[0]  # type: ignore
                assert (
                    stats["app"] == "no_typer" if env is no_typer else "with_typer"
                ), "Wrong app detected"
                assert stats["test_arg"] == 5

            cmd = [django_admin, *cmd]
            if hlp:
                cmd.append("--help")
                run_command(*cmd, **env)  # make sure byte code is compiled

            import_trace(*cmd, **env)  # warm up
            imports = import_trace(
                *cmd,
                save=(
                    f"{run.cmd}_"
                    f"{'typer' if run.typer else 'base'}_"
                    f"{'app_' if run.app else ''}"
                    f"{'rich_' if run.rich else ''}"
                    f"{'help' if run.help else ''}.log"
                ),
                **env,
            )
            total_import_seconds = (
                float(sum([imprt[0] for imprt in imports.values()])) / 10**6
            )

            times = []
            for _ in range(run_counts):
                times.append(time_command(*cmd, **env)[-1])

            profile_data[pos_str]["runs"].setdefault(hash(run), {})
            profile_data[pos_str]["runs"][hash(run)] = {
                "run": run,
                "time": sum(times) / run_counts,
                "import_time": total_import_seconds,
                "modules": len(imports),
            }

    with open(profile_file, "w") as f:
        json.dump(profile_data, f, indent=4, default=dataclass_encoder)

    pprint(profile_data[pos_str])


@app.command(help="Update documentation to reflect latest profiling data.")
def document():
    if not rich_installed:
        raise typer.Exit("rich must be installed to run document")

    from rich import box
    from rich.console import Console
    from rich.table import Table
    from rich.text import Text

    profile = load()
    key = max((int(k) for k in profile.keys()))
    current = profile[str(key)]

    def to_str(ver):
        return ".".join((str(v) for v in ver))

    def minimal_table():
        console = Console(record=True, width=80)

        # Top "header"
        console.print(
            Text(
                f"{current['platform'].os} | {current['platform'].arch} | "
                f"Python {to_str(current['env'].python)}",
                style="bold magenta",
            ),
            justify="center",
        )
        console.print(
            Text(
                f"django-typer {to_str(current['env'].django_typer)} | Typer {to_str(current['env'].typer)}"
                f" | Django {to_str(current['env'].django)} ",
                style="dim",
            ),
            justify="center",
        )

        table = Table(
            show_header=True,
            box=box.SIMPLE_HEAVY,
            expand=True,
        )

        # Row-label column
        table.add_column("", style="bold white on grey23")

        no_rich_cols = [
            (
                "BaseCommand",
                RunKey(cmd="minimal", typer=False, app=False, rich=False, help=False),
            ),
            (
                "TyperCommand",
                RunKey(cmd="minimal", typer=True, app=True, rich=False, help=False),
            ),
        ]
        rich_cols = [
            (
                "TyperCommand (rich)",
                RunKey(cmd="minimal", typer=True, app=True, rich=True, help=False),
            ),
            (
                "TyperCommand --help (rich)",
                RunKey(cmd="minimal", typer=True, app=True, rich=True, help=True),
            ),
        ]

        no_rich_style = {
            "style": "black on light_goldenrod2",
            "header_style": "bold black on light_goldenrod2",
        }
        rich_style = {
            "style": "black on light_sky_blue1",
            "header_style": "bold black on light_sky_blue1",
        }
        for name, _ in no_rich_cols:
            table.add_column(name, justify="center", **no_rich_style)
        for name, _ in rich_cols:
            table.add_column(name, justify="center", **rich_style)

        for label, values in [
            (
                "modules",
                [
                    str(current["runs"][str(hash(key))]["modules"])
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
            (
                "time",
                [
                    f"{current['runs'][str(hash(key))]['time']:.2f}"
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
            (
                "imports",
                [
                    f"{current['runs'][str(hash(key))]['import_time']:.2f}"
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
        ]:
            table.add_row(label, *values)

        console.print(table)

        svg = console.export_svg(title="CLI Startup Benchmarks")  # export as a string
        (
            Path(__file__).parent.parent / "doc/source/_static/img/minimal_profile.svg"
        ).write_text(svg, encoding="utf-8")

        console.print(table)

        svg = console.export_text()  # export as a string
        (
            Path(__file__).parent.parent / "doc/source/_static/img/minimal_profile.txt"
        ).write_text(svg, encoding="utf-8")

    def polls_table():
        console = Console(record=True, width=100)

        # Top "header"
        console.print(
            Text(
                f"{current['platform'].os} | {current['platform'].arch} | "
                f"Python {to_str(current['env'].python)}",
                style="bold magenta",
            ),
            justify="center",
        )
        console.print(
            Text(
                f"django-typer {to_str(current['env'].django_typer)} | Typer {to_str(current['env'].typer)}"
                f" | Django {to_str(current['env'].django)} ",
                style="dim",
            ),
            justify="center",
        )

        table = Table(
            show_header=True,
            box=box.SIMPLE_HEAVY,
            expand=True,
        )

        # Row-label column
        table.add_column("", style="bold white on grey23")

        no_rich_cols = [
            (
                "polls (tutorial)",
                RunKey(cmd="polls", typer=False, app=False, rich=False, help=False),
            ),
            (
                "polls --help (tutorial)",
                RunKey(cmd="polls", typer=False, app=False, rich=False, help=True),
            ),
            (
                "polls (typer)",
                RunKey(cmd="polls", typer=True, app=True, rich=False, help=False),
            ),
            (
                "shell completion",
                RunKey(
                    cmd="shellcompletion", typer=True, app=True, rich=False, help=False
                ),
            ),
        ]
        rich_cols = [
            (
                "polls (typer w/rich)",
                RunKey(cmd="polls", typer=True, app=True, rich=True, help=False),
            ),
            (
                "polls --help (typer w/rich)",
                RunKey(cmd="polls", typer=True, app=True, rich=True, help=True),
            ),
            (
                "shell completion (rich)",
                RunKey(
                    cmd="shellcompletion", typer=True, app=True, rich=True, help=False
                ),
            ),
        ]

        no_rich_style = {
            "style": "black on light_goldenrod2",
            "header_style": "bold black on light_goldenrod2",
        }
        rich_style = {
            "style": "black on light_sky_blue1",
            "header_style": "bold black on light_sky_blue1",
        }
        for name, _ in no_rich_cols:
            table.add_column(name, justify="center", **no_rich_style)
        for name, _ in rich_cols:
            table.add_column(name, justify="center", **rich_style)

        for label, values in [
            (
                "modules",
                [
                    str(current["runs"][str(hash(key))]["modules"])
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
            (
                "time",
                [
                    f"{current['runs'][str(hash(key))]['time']:.2f}"
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
            (
                "imports",
                [
                    f"{current['runs'][str(hash(key))]['import_time']:.2f}"
                    for _, key in (no_rich_cols + rich_cols)
                ],
            ),
        ]:
            table.add_row(label, *values)

        console.print(table)

        svg = console.export_svg(title="CLI Startup Benchmarks")  # export as a string
        (
            Path(__file__).parent.parent / "doc/source/_static/img/polls_profile.svg"
        ).write_text(svg, encoding="utf-8")

        console.print(table)
        svg = console.export_text()  # export as a string
        (
            Path(__file__).parent.parent / "doc/source/_static/img/polls_profile.txt"
        ).write_text(svg, encoding="utf-8")

    minimal_table()
    polls_table()


if __name__ == "__main__":
    app()  # type: ignore



from django.core.management import BaseCommand


class Command(BaseCommand):
    requires_system_checks = []
    requires_migrations_checks = False

    def add_arguments(self, parser) -> None:
        parser.add_argument("test_arg", type=int)
        parser.add_argument("--print", action="store_true")
        return super().add_arguments(parser)

    def handle(self, test_arg, print, **_):
        if print:
            import json

            return json.dumps(
                {
                    "app": "no_typer",
                    "test_arg": test_arg,
                }
            )



"""
Common types for command line argument specification.
"""

import sys
from pathlib import Path
from typing import Annotated, Optional, cast

from django.core.management import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Option

from .completers.path import directories, import_paths

COMMON_PANEL = "Django"


def print_version(context, _, value):
    """
    A callback to run the :meth:`~django.core.management.BaseCommand.get_version` routine
    of the command when --version is specified.
    """
    if value:
        context.django_command.stdout.write(context.django_command.get_version())
        sys.exit()


def set_no_color(context, _, value):
    """
    If the value was provided set it on the command.
    """
    if value:
        context.django_command.no_color = value
        if context.params.get("force_color", False):
            raise CommandError(
                "The --no-color and --force-color options can't be used together."
            )
    return value


def set_force_color(context, _, value):
    """
    If the value was provided set it on the command.
    """
    if value:
        context.django_command.force_color = value
    return value


def show_locals(context, param, _):
    from click.core import ParameterSource

    if context.get_parameter_source(param.name) is not ParameterSource.DEFAULT:
        from .config import traceback_config
        from .utils import install_traceback

        install_traceback(
            {
                **traceback_config(),
                "show_locals": param.name == "show_locals",
            }
        )


Version = Annotated[
    bool,
    Option(
        "--version",
        help=cast(str, _("Show program's version number and exit.")),
        callback=print_version,
        is_eager=True,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :meth:`--version <django.core.management.BaseCommand.get_version>`
option.

The ``--version`` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand`.
"""


Verbosity = Annotated[
    int,
    Option(
        help=cast(
            str,
            _(
                "Verbosity level; 0=minimal output, 1=normal output, "
                "2=verbose output, 3=very verbose output"
            ),
        ),
        show_choices=True,
        min=0,
        max=3,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :option:`--verbosity` option.
:class:`~django_typer.management.TyperCommand` does not include the verbosity option by
default, but it can be added to the command like so if needed.

.. code-block:: python

        from django_typer.types import Verbosity

        def handle(self, verbosity: Verbosity = 1):
            ...
"""

Settings = Annotated[
    str,
    Option(
        help=cast(
            str,
            _(
                "The Python path to a settings module, e.g. "
                '"myproject.settings.main". If this isn\'t provided, the '
                "DJANGO_SETTINGS_MODULE environment variable will be used."
            ),
        ),
        rich_help_panel=COMMON_PANEL,
        shell_complete=import_paths,
        show_default=False,
    ),
]
"""
The type hint for the :option:`--settings` option.

The :option:`--settings` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to specify or override the settings
module to use.
"""


PythonPath = Annotated[
    Optional[Path],
    Option(
        help=cast(
            str,
            _(
                "A directory to add to the Python path, e.g. "
                '"/home/djangoprojects/myproject".'
            ),
        ),
        rich_help_panel=COMMON_PANEL,
        shell_complete=directories,
        show_default=False,
    ),
]
"""
The type hint for the :option:`--pythonpath` option.

The :option:`--pythonpath` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to specify a directory to add to the
Python sys path.
"""


Traceback = Annotated[
    bool,
    Option(
        "--traceback",
        help=cast(str, _("Raise on CommandError exceptions")),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :option:`--traceback` option.

The :option:`--traceback` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to allow
:class:`~django.core.management.CommandError` exceptions to propagate out of the command
and produce a stack trace.
"""


ShowLocals = Annotated[
    bool,
    Option(
        "--show-locals",
        help=cast(
            str,
            _("Print local variables in tracebacks."),
        ),
        callback=show_locals,
        is_eager=True,
        rich_help_panel=COMMON_PANEL,
        show_default=False,
    ),
]
"""
A toggle to turn on exception traceback local variable rendering in rich
tracebacks.
"""

HideLocals = Annotated[
    bool,
    Option(
        "--hide-locals",
        help=cast(str, _("Hide local variables in tracebacks.")),
        callback=show_locals,
        is_eager=True,
        rich_help_panel=COMMON_PANEL,
        show_default=False,
    ),
]
"""
A toggle to turn off exception traceback local variable rendering in rich
tracebacks.
"""

NoColor = Annotated[
    bool,
    Option(
        "--no-color",
        help=cast(str, _("Don't colorize the command output.")),
        is_eager=True,
        callback=set_no_color,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :option:`--no-color` option.

The :option:`--no-color` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to force disable colorization of the
command. You can check the supplied value of :option:`--no-color` by checking the
no_color attribute of the command instance.
"""

ForceColor = Annotated[
    bool,
    Option(
        "--force-color",
        help=cast(str, _("Force colorization of the command output.")),
        is_eager=True,
        callback=set_force_color,
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :option:`--force-color` option.

The :option:`--force-color` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to force colorization of the
command. You can check the supplied value of :option:`--force-color` by checking the
force_color attribute of the command instance.
"""

SkipChecks = Annotated[
    bool,
    Option(
        "--skip-checks",
        help=cast(str, _("Skip system checks.")),
        rich_help_panel=COMMON_PANEL,
    ),
]
"""
The type hint for the :option:`--skip-checks` option.

The :option:`--skip-checks` option is included by default and behaves the same as on
:class:`~django.core.management.BaseCommand` use it to skip system checks.
"""



import typing as t

from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import TyperCommand, command, group


class Command(TyperCommand):
    help = _("Test multiple groups commands and callbacks")

    precision = 2
    verbosity = 1

    @command()
    def echo(self, message: str) -> str:
        """
        Echo the given message.
        """
        assert issubclass(self.__class__, Command)
        return message

    @group()
    def math(self, precision: int = precision):
        """
        Do some math at the given precision.
        """
        assert issubclass(self.__class__, Command)
        self.precision = precision

    @math.command()
    def multiply(
        self,
        number1: t.Annotated[
            float, Argument(help=_("The first number."), show_default=False)
        ],
        number2: t.Annotated[
            float, Argument(help=_("The second number."), show_default=False)
        ],
        numbers: t.Annotated[
            t.List[float],
            Argument(
                help=_("The list of numbers to multiply: n1*n2*n3*...*nN. "),
                show_default=False,
            ),
        ],
    ):
        """
        Multiply the given numbers.
        """
        assert issubclass(self.__class__, Command)
        res = number1 * number2
        for n in numbers:
            res *= n
        return f"{res:.{self.precision}f}"

    @math.command()
    def divide(
        self,
        number1: t.Annotated[
            float, Argument(help=_("The numerator."), show_default=False)
        ],
        number2: t.Annotated[
            float, Argument(help=_("The denominator."), show_default=False)
        ],
        numbers: t.Annotated[
            t.List[float],
            Argument(
                help=_("Additional denominators: n1/n2/n3/.../nN."), show_default=False
            ),
        ],
    ) -> str:
        """
        Divide the given numbers.
        """
        assert issubclass(self.__class__, Command)
        res = number1 / number2
        for n in numbers:
            res /= n
        return f"{res:.{self.precision}f}"

    @group()
    def string(
        self,
        string: t.Annotated[
            str, Argument(help=_("The string to operate on."), show_default=False)
        ],
    ):
        """
        String operations.
        """
        assert issubclass(self.__class__, Command)
        self.op_string = string

    @string.group()
    def case(self):
        """
        Case operations.
        """
        assert issubclass(self.__class__, Command)

    @case.command()
    def upper(
        self,
        begin: t.Annotated[
            int, Argument(help=_("The starting index of the string to operate on."))
        ] = 0,
        end: t.Annotated[
            t.Optional[int],
            Argument(help=_("The ending index of the string to operate on.")),
        ] = None,
    ) -> str:
        """
        Convert the given string to upper case.
        """
        assert issubclass(self.__class__, Command)
        return f"{self.op_string[0:begin]}{self.op_string[begin:end].upper()}{self.op_string[end:None] if end else ''}"

    @case.command()
    def lower(
        self,
        begin: t.Annotated[
            int, Option(help=_("The starting index of the string to operate on."))
        ] = 0,
        end: t.Annotated[
            t.Optional[int],
            Option(help=_("The ending index of the string to operate on.")),
        ] = None,
    ):
        """
        Convert the given string to upper case.
        """
        assert issubclass(self.__class__, Command)
        return f"{self.op_string[0:begin]}{self.op_string[begin:end].lower()}{self.op_string[end:None] if end else ''}"

    @string.command()
    def split(self, sep: str = " "):
        """
        Split the given string on the given separator.
        """
        assert issubclass(self.__class__, Command)
        return " ".join(self.op_string.split(sep))

    def test(self, a: int, b: int) -> int:
        return a + b



from typing import Annotated

from django.core.management import CommandError
from typer import Option

from django_typer.management import TyperCommand, command
from django_typer import types
from tests.utils import log_django_parameters


class Command(TyperCommand):
    help = "Test that django parameters work as expected"

    @command(epilog="epilog")
    def handle(
        self,
        throw: bool = False,
        verbosity: types.Verbosity = 1,
        traceback: Annotated[
            bool,
            Option(
                help=("Raise on CommandError exceptions"),
                rich_help_panel=types.COMMON_PANEL,
            ),
        ] = True,  # this should change the default!
    ):
        assert self.__class__ is Command
        log_django_parameters(self, verbosity=verbosity, traceback=traceback)
        if throw:
            raise CommandError("Test Exception")



from django.core.management import BaseCommand, CommandError

from profiling.polls.models import Question as Poll


class Command(BaseCommand):
    help = "Closes the specified poll for voting"

    def add_arguments(self, parser):
        parser.add_argument("poll_ids", nargs="+", type=int)

        # Named (optional) arguments
        parser.add_argument(
            "--delete",
            action="store_true",
            help="Delete poll instead of closing it",
        )

    def handle(self, *args, **options):
        for poll_id in options["poll_ids"]:
            try:
                poll = Poll.objects.get(pk=poll_id)
            except Poll.DoesNotExist:
                raise CommandError('Poll "%s" does not exist' % poll_id)

            poll.opened = False
            poll.save()

            self.stdout.write(
                self.style.SUCCESS('Successfully closed poll "%s"' % poll_id)
            )

            if options["delete"]:
                poll.delete()



from django.apps import AppConfig

from django_typer.utils import register_command_plugins


class BackupFilesConfig(AppConfig):
    name = "tests.apps.examples.plugins.backup_files"
    label = name.replace(".", "_")



from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import run_command


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter0",
        "tests.apps.test_app2",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class InterferenceTests(TestCase):
    """
    commands can be adapted by extending or overriding them. However, when done
    this should *not* affect the root command. These tests test that the root
    command can be imported and used directly even when it is extended by apps
    further up the app stack.

    TODO - if we can fix the argument name interference problem, this would be a
    good test to also test that - just rename all the arguments in the command to arg#.
    """

    def test_inherit_command_interfence(self):
        self.assertEqual(
            call_command("interference", "cmd1", "abc"),
            "test_app::interference::cmd1(abc)",
        )
        self.assertEqual(
            call_command("interference_ext", "cmd1", "5"),
            "test_app2::interference::cmd1(5)",
        )

        self.assertEqual(
            call_command("interference", "cmd2", "10"),
            "test_app::interference::cmd2(10)",
        )
        self.assertEqual(
            call_command("interference_ext", "cmd2", "10"),
            "test_app::interference::cmd2(10)",
        )

        self.assertEqual(
            call_command("interference_ext", "cmd-ext"),
            "test_app2::interference::cmd_ext()",
        )
        with self.assertRaises(CommandError):
            call_command("interference", "cmd-ext")

        with self.assertRaises(CommandError):
            call_command("interference", "adapt-cmd")

        self.assertEqual(
            call_command("interference_ext", "adapt-cmd"),
            "test_app2::interference::adapt_cmd()",
        )

    def test_inherit_group_interfence(self):
        self.assertEqual(
            call_command("interference", "grp1", "1.2"),
            "test_app::interference::grp1(1.2)",
        )
        self.assertEqual(
            call_command("interference", "grp1", "1.3", "cmd3", "abc", "def"),
            "test_app::interference::grp1(1.3)::cmd3(abc, def)",
        )
        self.assertEqual(
            call_command("interference", "grp1", "1.4", "cmd4", "3", "7"),
            "test_app::interference::grp1(1.4)::cmd4(3, 7)",
        )
        self.assertEqual(
            call_command("interference", "grp1", "1.4", "subgroup", "--arg5-0"),
            "test_app::interference::grp1(1.4)::subgroup(True)",
        )
        self.assertEqual(
            call_command("interference", "grp1", "1.4", "subgroup", "--arg5-0"),
            "test_app::interference::grp1(1.4)::subgroup(True)",
        )
        self.assertEqual(
            call_command(
                "interference",
                "grp1",
                "1.4",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            ),
            "test_app::interference::grp1(1.4)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        with self.assertRaises(CommandError):
            call_command("interference", "grp2", "xyz")

        with self.assertRaises(CommandError):
            call_command("interference", "grp1", "1.4", "subgroup", "--arg5-0", "cmd5")

        with self.assertRaises(CommandError):
            call_command(
                "interference",
                "grp1",
                "1.4",
                "subgroup",
                "subsubgroup",
                "subsubcommand2",
            )

        with self.assertRaises(CommandError):
            call_command("interference", "adapt-group")

        with self.assertRaises(CommandError):
            call_command("interference", "subgroup")

        with self.assertRaises(CommandError):
            call_command("interference", "subsubgroup")

        self.assertEqual(
            call_command("interference_ext", "grp1", "1.2"),
            "test_app::interference::grp1(1.2)",
        )
        self.assertEqual(
            call_command("interference_ext", "grp1", "2.5", "cmd3", "8", "9"),
            "test_app2::interference::grp1(2.5)::cmd3(8, 9)",
        )
        self.assertEqual(
            call_command("interference_ext", "grp1", "1.4", "cmd4", "3", "7"),
            "test_app::interference::grp1(1.4)::cmd4(3, 7)",
        )
        self.assertEqual(
            call_command("interference_ext", "grp2", "xyz"),
            "test_app2::interference::grp2(xyz)",
        )
        self.assertEqual(
            call_command(
                "interference_ext", "grp1", "1.4", "subgroup", "--arg5-0", "cmd5"
            ),
            "test_app2::interference::grp1(1.4)::subgroup(True)::cmd5()",
        )
        self.assertEqual(
            call_command(
                "interference_ext",
                "grp1",
                "1.4",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            ),
            "test_app2::interference::grp1(1.4)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        self.assertEqual(
            call_command(
                "interference_ext",
                "grp1",
                "1.4",
                "subgroup",
                "subsubgroup",
                "subsubcommand2",
            ),
            "test_app2::interference::grp1(1.4)::subgroup(False)::subsubgroup(True)::subsubcommand2()",
        )
        self.assertEqual(
            call_command("interference_ext", "adapt-group"),
            "test_app2::interference::adapt_group()",
        )

        with self.assertRaises(CommandError):
            call_command("interference_ext", "subgroup")

        with self.assertRaises(CommandError):
            call_command("interference_ext", "subsubgroup")

    def test_inherit_initialize_interference(self):
        self.assertEqual(
            call_command("interference"), "test_app::interference::init(True)"
        )
        self.assertEqual(
            call_command("interference_ext"), "test_app2::interference::init(True)"
        )

    def test_initialize_classmethod_run(self):
        stdout, stderr, returncode = run_command(
            "interference_ext_init",
            "--settings",
            "tests.settings.override",
        )
        self.assertEqual(returncode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "test_app2::interference_ext_init::init(5)")

        stdout, stderr, returncode = run_command(
            "method_override",
            "--settings",
            "tests.settings.adapted",
        )
        self.assertEqual(returncode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "adapter0::init()")

    def test_initialize_classmethod_call(self):
        self.assertEqual(
            call_command("interference_ext_init"),
            "test_app2::interference_ext_init::init(5)",
        )

        self.assertEqual(
            call_command("method_override"),
            "adapter0::init()",
        )

    def test_initialize_classmethod_direct(self):
        from django_typer.management import TyperCommand

        command = get_command("interference_ext_init", TyperCommand)
        self.assertEqual(command.init(), "test_app2::interference_ext_init::init(5)")

        command = get_command("method_override", TyperCommand)
        self.assertEqual(command.init(), "adapter0::init()")

    def test_command_classmethod_run(self):
        stdout, stderr, returncode = run_command(
            "method_override",
            "--settings",
            "tests.settings.adapted",
            "cmd1",
        )
        self.assertEqual(returncode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "adapter0::cmd1()")

        stdout, stderr, returncode = run_command(
            "method_override",
            "--settings",
            "tests.settings.adapted",
            "cmd2",
        )
        self.assertEqual(returncode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "adapter0::cmd2()")

    def test_command_classmethod_call(self):
        self.assertEqual(
            call_command("method_override", "cmd1"),
            "adapter0::cmd1()",
        )
        self.assertEqual(
            call_command("method_override", "cmd2"),
            "adapter0::cmd2()",
        )

    def test_command_classmethod_direct(self):
        from django_typer.management import TyperCommand

        command = get_command("method_override", TyperCommand)
        self.assertEqual(command.cmd1(), "adapter0::cmd1()")
        self.assertEqual(command.cmd2(), "adapter0::cmd2()")



from django.core.management import CommandError

from django_typer.management import TyperCommand, command, initialize
from django_typer import types
from tests.utils import log_django_parameters


class Command(TyperCommand):
    help = "Test that django parameters work as expected"

    @initialize()
    def init(self, verbosity: types.Verbosity = 1):
        """
        The callback to initialize the command.
        """
        assert self.__class__ is Command
        self.verbosity = verbosity

    @command()
    def cmd1(self, throw: bool = False):
        assert self.__class__ is Command
        log_django_parameters(self, verbosity=self.verbosity)
        if throw:
            raise CommandError("Test Exception")

    @command()
    def cmd2(self, throw: bool = False):
        assert self.__class__ is Command
        log_django_parameters(self, verbosity=self.verbosity)
        if throw:
            raise CommandError("Test Exception")



import sys
import typing as t
from pathlib import Path

from tests.apps.examples.plugins.backup_files.management.commands.backup_typer import (
    plugins,
    hookimpl,
)


@hookimpl
def backup_files(command) -> t.Optional[Path]:
    # this is where you would put your custom file backup logic
    return command.output_directory / "files2.zip"


plugins.register(sys.modules[__name__])



import pytest
from pathlib import Path
import inspect
import os
import sys
from django import VERSION
from packaging.version import parse as parse_version


# conftest.py
def pytest_collection_modifyitems(items):
    """Modifies test items in place to ensure plugin tests run in a specific order."""
    sorted_tests = []
    interference_tests = []
    plugin_tests = []  # push these to the back
    native_plugin_tests = []

    for test in items:
        if "plugin_pattern" in test.module.__name__:
            plugin_tests.append(test)
        elif "native_plugin" in test.module.__name__:
            native_plugin_tests.append(test)
        elif "test_interference" in test.module.__name__:
            interference_tests.append(test)
        else:
            sorted_tests.append(test)

    items[:] = [
        *sorted_tests,
        *plugin_tests,
        *native_plugin_tests,
        *interference_tests,
    ]


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    log_file = Path(__file__).parent / "tests.log"
    outcome = yield
    report = outcome.get_result()
    if report.when == "call" and report.outcome == "passed":
        if log_file.exists():
            with open(log_file, "a") as log_file:
                log_file.write(f"{item.nodeid}\n")


def first_breakable_line(obj) -> tuple[str, int]:
    """
    Return the absolute line number of the first executable statement
    in a function or bound method.
    """
    import ast
    import textwrap

    func = obj.__func__ if inspect.ismethod(obj) else obj

    source = inspect.getsource(func)
    source = textwrap.dedent(source)
    filename = inspect.getsourcefile(func)
    assert filename
    _, start_lineno = inspect.getsourcelines(func)

    tree = ast.parse(source)

    for node in tree.body[0].body:
        if (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and isinstance(node.value.value, str)
        ):
            continue

        return filename, start_lineno + node.lineno - 1

    # fallback: just return the line after the def
    return filename, start_lineno + 1


def pytest_runtest_call(item):
    # --trace cli option does not work for unittest style tests so we implement it here
    test = getattr(item, "obj", None)
    if item.config.option.trace and inspect.ismethod(test):
        from IPython.terminal.debugger import TerminalPdb

        try:
            file = inspect.getsourcefile(test)
            assert file
            dbg = TerminalPdb()
            dbg.set_break(*first_breakable_line(test))
            dbg.cmdqueue.append("continue")
            dbg.set_trace()
        except (OSError, AssertionError):
            pass


def pytest_configure(config: pytest.Config) -> None:

    if os.getenv("GITHUB_ACTIONS") == "true":
        # verify that the environment is set up correctly - this is used in CI to make
        # sure we're testing against the dependencies we think we are
        expected_python = os.getenv("TEST_PYTHON_VERSION")
        expected_django = os.getenv("TEST_DJANGO_VERSION")

        if expected_python:
            expected_python = parse_version(expected_python)
            if sys.version_info[:2] != (expected_python.major, expected_python.minor):
                raise pytest.UsageError(
                    f"Python Version Mismatch: {sys.version_info[:2]} != "
                    f"{expected_python}"
                )

        if expected_django:
            dj_actual = VERSION[:2]
            expected_django = parse_version(expected_django)
            dj_expected = (expected_django.major, expected_django.minor)
            if dj_actual != dj_expected:
                raise pytest.UsageError(
                    f"Django Version Mismatch: {dj_actual} != {expected_django}"
                )



import json

from django_typer.management import TyperCommand, initialize


class Command(TyperCommand):
    help = "Test basic callback command."

    parameters = {}

    @initialize()
    def init(self, p1: int, flag1: bool = False, flag2: bool = True):
        """
        The callback to initialize the command.
        """
        assert self.__class__ is Command
        self.parameters = {"p1": p1, "flag1": flag1, "flag2": flag2}

    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        self.parameters.update({"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4})
        return json.dumps(self.parameters)



from django_typer.management import TyperCommand, command, group, initialize
from .inheritance2_1 import Command as Inheritance21Command
from .inheritance2_2 import Command as Inheritance22Command


class Command(Inheritance21Command, Inheritance22Command):
    @command()
    def b(self):
        assert isinstance(self, Command)
        return "inheritance3::b()"

    @command()
    def d(self):
        assert isinstance(self, Command)
        return "inheritance3::d()"

    @Inheritance22Command.g2.command()
    def g2a(self):
        assert isinstance(self, Command)
        return "inheritance3::g2::g2a()"



import typing as t
import json
import typer
from django_typer.management import Typer
from django_typer.parsers.model import ModelObjectParser, ReturnType
from django_typer.completers import these_strings
from django.db.models import QuerySet

from tests.apps.test_app.models import ChoicesShellCompleteTester


app = Typer()


def model_parser(field: str) -> ModelObjectParser:
    return ModelObjectParser(
        ChoicesShellCompleteTester,
        field,
        return_type=ReturnType.QUERY_SET,
    )


@app.command()
def choices(
    char_choices: t.Annotated[
        QuerySet[ChoicesShellCompleteTester],
        typer.Option(
            parser=model_parser("char_choice"),
            shell_complete=these_strings(
                ChoicesShellCompleteTester.CHAR_CHOICES,
                allow_duplicates=False,
            ),
            help="Fetch objects by their char choice fields.",
        ),
    ] = ChoicesShellCompleteTester.objects.none(),
    int_choices: t.Annotated[
        QuerySet[ChoicesShellCompleteTester],
        typer.Option(
            parser=model_parser("int_choice"),
            shell_complete=these_strings(
                ChoicesShellCompleteTester.INT_CHOICES,
                allow_duplicates=False,
            ),
            help="Fetch objects by their int choice fields.",
        ),
    ] = ChoicesShellCompleteTester.objects.none(),
    ip_choices: t.Annotated[
        QuerySet[ChoicesShellCompleteTester],
        typer.Option(
            parser=model_parser("ip_choice"),
            shell_complete=these_strings(
                ChoicesShellCompleteTester.IP_CHOICES,
                allow_duplicates=False,
            ),
            help="Fetch objects by their ip choice fields.",
        ),
    ] = ChoicesShellCompleteTester.objects.none(),
):
    objects = {
        "char_choice": {},
        "int_choice": {},
        "ip_choice": {},
    }
    for choice_field, record in objects.items():
        for choice in locals()[f"{choice_field}"] or []:
            assert isinstance(choice, ChoicesShellCompleteTester)
            record.setdefault(getattr(choice, choice_field), [])
            record[getattr(choice, choice_field)].append(choice.pk)

    return json.dumps(objects)



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class EchoWrapperTests(TestCase):
    def test_echo_no_color(self):
        result = run_command(
            "echo", "--no-color", "echo-test", "Brian Kohan", "--color", "magenta"
        )[0]
        self.assertTrue("\x1b[35m" not in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        result = run_command(
            "echo",
            "--no-color",
            "echo-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            "--no-nl",
        )[1]
        self.assertTrue("\x1b[35m" not in result)
        self.assertTrue("\n" not in result)
        self.assertTrue("Brian Kohan" in result)

    def test_echo_force_color(self):
        result = run_command(
            "echo", "--force-color", "echo-test", "Brian Kohan", "--color", "magenta"
        )[0]
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        result = run_command(
            "echo",
            "--force-color",
            "echo-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            "--no-nl",
        )[1]
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" not in result)
        self.assertTrue("Brian Kohan" in result)

    def test_echo_call_command(self):
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        call_command(
            "echo",
            "--force-color",
            "echo-test",
            "Brian Kohan",
            "--color",
            "magenta",
            stdout=stdout_buffer,
            stderr=stderr_buffer,
        )
        result = stdout_buffer.getvalue()
        stdout_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        call_command(
            "echo",
            "--force-color",
            "echo-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            stdout=stdout_buffer,
            stderr=stderr_buffer,
        )
        result = stderr_buffer.getvalue()
        stderr_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

    def test_echo_get_command(self):
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        echo = get_command(
            "echo", force_color=True, stdout=stdout_buffer, stderr=stderr_buffer
        )
        echo.echo_test("Brian Kohan", color="magenta")
        result = stdout_buffer.getvalue()
        stdout_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        echo.echo_test("Brian Kohan", color="magenta", error=True)
        result = stderr_buffer.getvalue()
        stderr_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

    def test_secho_no_color(self):
        result = run_command(
            "echo", "--no-color", "secho-test", "Brian Kohan", "--color", "magenta"
        )[0]
        self.assertTrue("\x1b[35m" not in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        result = run_command(
            "echo",
            "--no-color",
            "secho-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            "--no-nl",
        )[1]
        self.assertTrue("\x1b[35m" not in result)
        self.assertTrue("\n" not in result)
        self.assertTrue("Brian Kohan" in result)

    def test_secho_force_color(self):
        result = run_command(
            "echo", "--force-color", "secho-test", "Brian Kohan", "--color", "magenta"
        )[0]
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        result = run_command(
            "echo",
            "--force-color",
            "secho-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            "--no-nl",
        )[1]
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" not in result)
        self.assertTrue("Brian Kohan" in result)

    def test_secho_call_command(self):
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        call_command(
            "echo",
            "--force-color",
            "secho-test",
            "Brian Kohan",
            "--color",
            "magenta",
            stdout=stdout_buffer,
            stderr=stderr_buffer,
        )
        result = stdout_buffer.getvalue()
        stdout_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        call_command(
            "echo",
            "--force-color",
            "secho-test",
            "Brian Kohan",
            "--color",
            "magenta",
            "--error",
            stdout=stdout_buffer,
            stderr=stderr_buffer,
        )
        result = stderr_buffer.getvalue()
        stderr_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

    def test_secho_get_command(self):
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        echo = get_command(
            "echo", force_color=True, stdout=stdout_buffer, stderr=stderr_buffer
        )
        echo.secho_test("Brian Kohan", color="magenta")
        result = stdout_buffer.getvalue()
        stdout_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)

        echo.secho_test("Brian Kohan", color="magenta", error=True)
        result = stderr_buffer.getvalue()
        stderr_buffer.seek(0)
        self.assertTrue("\x1b[35m" in result)
        self.assertTrue("\n" in result)
        self.assertTrue("Brian Kohan" in result)



"""
Typer_ supports custom parsers for options and arguments. If you would
like to type a parameter with a type that isn't supported by Typer_ you can
`implement your own parser
<https://typer.tiangolo.com/tutorial/parameter-types/custom-types>`_, or
:class:`click.ParamType` in :doc:`click <click:index>` parlance.

This package contains a collection of parsers that turn strings into useful
Django types. Pass these parsers to the `parser` argument of ``typer.Option`` and
``typer.Argument``. Parsers are provided for:

- **models, querysets or field values**: Turn a string into a model object instance
    using :class:`~django_typer.parsers.model.ModelObjectParser`.
- **Django apps**: Turn a string into an AppConfig instance using
    :func:`~django_typer.parsers.apps.app_config`.


.. warning::

    If you implement a custom parser, please take care to ensure that it:
        - Handles the case where the value is already the expected type.
        - Returns None if the value is None (already implemented if subclassing
          ParamType).
        - Raises a :exc:`~django.core.management.CommandError` if the value is invalid.
        - Handles the case where the param and context are None.
"""



import os
from tests.utils import run_command
import time
import pytest
from pprint import pformat
from .utils import rich_installed
import platform


@pytest.mark.no_rich
@pytest.mark.skipif(
    rich_installed, reason="Rich should not be installed to test module bloat."
)
def test_performance_regression():
    env = dict(os.environ)

    # disable coverage
    for var in [
        "COVERAGE_PROCESS_START",
        "COV_CORE_SOURCE",
        "COV_CORE_CONFIG",
        "COV_CORE_DATAFILE",
        "PYTEST_XDIST_WORKER",
    ]:
        env.pop(var, None)

    start = time.perf_counter()
    result, stderr, retcode = run_command(
        "perf",
        "5",
        "--print",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_no_typer"},
    )
    end = time.perf_counter()
    if retcode:
        pytest.fail(stderr)
    assert result["print"]
    assert result["no_typer"] == 5
    mods_no_typer = result["modules"]
    no_typer_time = end - start

    start = time.perf_counter()
    result, stderr, retcode = run_command(
        "perf",
        "5",
        "--print",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_typer"},
    )
    end = time.perf_counter()
    if retcode:
        pytest.fail(stderr)
    assert result["print"]
    assert result["typer"] == 5
    mods_typer = result["modules"]
    typer_time = end - start

    start = time.perf_counter()
    result, stderr, retcode = run_command(
        "perf",
        "5",
        "--print",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_typer_no_app"},
    )
    end = time.perf_counter()
    if retcode:
        pytest.fail(stderr)
    assert result["print"]
    assert result["typer"] == 5
    mods_typer_no_app = result["modules"]
    typer_no_app_time = end - start

    # print the stats
    print("\nWithout typer:\n\t")
    print(f"\ttime: {no_typer_time:0.4f}")
    print(f"\tmodules: {len(mods_no_typer)}")

    print("\nWith typer:\n\t")
    print(f"\ttime: {typer_time:0.4f}")
    print(f"\tmodules: {len(mods_typer)}")

    print("\nWith typer, but app not installed:\n\t")
    print(f"\ttime: {typer_no_app_time:0.4f}")
    print(f"\tmodules: {len(mods_typer_no_app)}")

    # notify us if adding typer inflates module count by more than 10 percent
    assert len(mods_no_typer) / len(mods_typer) > 0.9, (
        f"Typer modules added: \n{pformat(set(mods_typer) - set(mods_no_typer))}"
    )
    assert len(mods_no_typer) / len(mods_typer_no_app) > 0.9, (
        f"Typer modules added: \n{pformat(set(mods_typer_no_app) - set(mods_no_typer))}"
    )

    # notify us if adding typer inflates command exec time by more than 20 percent
    assert no_typer_time / typer_time > 0.2
    assert no_typer_time / typer_no_app_time > 0.2


@pytest.mark.no_rich
@pytest.mark.skipif(
    rich_installed, reason="Rich should not be installed to test module bloat."
)
@pytest.mark.skipif(platform.system() != "Darwin", reason="Test is only for macOS")
def test_timing():
    env = dict(os.environ)
    # disable coverage
    for var in [
        "COVERAGE_PROCESS_START",
        "COV_CORE_SOURCE",
        "COV_CORE_CONFIG",
        "COV_CORE_DATAFILE",
        "PYTEST_XDIST_WORKER",
    ]:
        env.pop(var, None)

    result, stderr, retcode, no_typer_seconds = run_command(
        "perf",
        "5",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_no_typer"},
        time=True,
    )
    if retcode:
        pytest.fail(stderr)
    assert not result, "perf should not have printed"

    result, stderr, retcode, typer_seconds = run_command(
        "perf",
        "5",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_typer"},
        time=True,
    )
    if retcode:
        pytest.fail(stderr)
    assert not result, "perf should not have printed"

    result, stderr, retcode, typer_no_app_seconds = run_command(
        "perf",
        "5",
        env={**env, "DJANGO_SETTINGS_MODULE": "tests.settings.perf_typer_no_app"},
        time=True,
    )
    if retcode:
        pytest.fail(stderr)
    assert not result, "perf should not have printed"

    # print the stats
    print("\nWithout typer:\n\t")
    print(f"\ttime: {no_typer_seconds:0.4f}")

    print("\nWith typer:\n\t")
    print(f"\ttime: {typer_seconds:0.4f}")

    print("\nWith typer, but app not installed:\n\t")
    print(f"\ttime: {typer_no_app_seconds:0.4f}")

    # notify us if adding typer inflates command exec time by more than 20 percent
    assert no_typer_seconds / typer_seconds > 0.2
    assert no_typer_seconds / typer_no_app_seconds > 0.2



import sys
from io import StringIO

import pytest
from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import rich_installed, run_command, similarity

adapted_help = """
Usage: ./manage.py adapted [OPTIONS] COMMAND [ARGS]...

Options:
  --version          Show program's version number and exit.
  --settings TEXT    The Python path to a settings module, e.g.
                     "myproject.settings.main". If this isn't provided, the
                     DJANGO_SETTINGS_MODULE environment variable will be used.
  --pythonpath PATH  A directory to add to the Python path, e.g.
                     "/home/djangoprojects/myproject".
  --traceback        Raise on CommandError exceptions
  --no-color         Don't colorize the command output.
  --force-color      Force colorization of the command output.
  --skip-checks      Skip system checks.
  --help             Show this message and exit.

Commands:
  adapted
  echo     Echo both messages.
  no-self
"""

adapted_adapted_help = """
Usage: ./manage.py adapted adapted [OPTIONS] MESSAGE

Arguments:
  MESSAGE  [required]

Options:
  --help  Show this message and exit.
"""

adapted_rich_help = """
 Usage: ./manage.py adapted [OPTIONS] COMMAND [ARGS]...                         
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ adapted                                                                      │
│ echo      Echo both messages.                                                │
│ no-self                                                                      │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

adapted_adapted_rich_help = """                           
 Usage: ./manage.py adapted adapted [OPTIONS] MESSAGE                           
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    message      TEXT  [required]                                           │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

adapted_echo_help = """
Usage: ./manage.py adapted echo [OPTIONS] MSG1 MSG2

  Echo both messages.

Arguments:
  MSG1  [required]
  MSG2  [required]

Options:
  --help  Show this message and exit.
"""

adapted_echo_rich_help = """
 Usage: ./manage.py adapted echo [OPTIONS] MSG1 MSG2                            
                                                                                
 Echo both messages.                                                            
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    msg1      TEXT  [required]                                              │
│ *    msg2      TEXT  [required]                                              │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

adapted_help_no_adapters = """
 Usage: ./manage.py adapted [OPTIONS] MESSAGE                                   
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    message      TEXT  [required]                                           │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
"""
adapted_help_no_adapters_no_rich = """
Usage: ./manage.py adapted [OPTIONS] MESSAGE

Arguments:
  MESSAGE  [required]

Options:
  --version          Show program's version number and exit.
  --settings TEXT    The Python path to a settings module, e.g.
                     "myproject.settings.main". If this isn't provided, the
                     DJANGO_SETTINGS_MODULE environment variable will be used.
  --pythonpath PATH  A directory to add to the Python path, e.g.
                     "/home/djangoprojects/myproject".
  --traceback        Raise on CommandError exceptions
  --no-color         Don't colorize the command output.
  --force-color      Force colorization of the command output.
  --skip-checks      Skip system checks.
  --help             Show this message and exit.
"""


class ResetAppsMixin:
    @classmethod
    def setUpClass(self):
        from django_typer import utils

        utils._command_extensions = {}
        for mod in list(sys.modules.keys()):
            if any(
                [
                    mod.startswith(path)
                    for path in [
                        "tests.apps.adapter1",
                        "tests.apps.adapter2",
                        "tests.apps.test_app.management.commandstests.apps.adapter0",
                    ]
                ]
            ):
                sys.modules[mod]
                del sys.modules[mod]
        super().setUpClass()


class UnadaptedTests(TestCase):
    @pytest.mark.rich
    @pytest.mark.skipif(not rich_installed, reason="rich is not installed")
    def test_no_adapter_apps(self):
        hlp = run_command("adapted", "--help", "--no-color")[0]
        self.assertGreater(
            sim := similarity(adapted_help_no_adapters, hlp),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted --help similiarity: {sim}")

    @pytest.mark.no_rich
    @pytest.mark.skipif(rich_installed, reason="rich is installed")
    def test_no_adapter_apps_no_rich(self):
        stdout = StringIO()
        adapted = get_command("adapted", no_color=True, stdout=stdout)
        adapted.print_help("./manage.py", "adapted")
        self.assertGreater(
            sim := similarity(adapted_help_no_adapters_no_rich, stdout.getvalue()),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted --help similiarity: {sim}")

    def test_adapted_is_compound(self):
        self.assertFalse(get_command("adapted").is_compound_command)


class Adapter2InterferenceTests(ResetAppsMixin, TestCase):
    """
    A very complicated test that adapts a command that is extended
    from another complicated command. This also insures adaptations
    do not unintentionally interfere with upstream base commands.
    """

    def test_adapted2_baseline(self):
        self.assertEqual(
            run_command(
                "adapted2", "--verbosity", "3", "grp2", "--flag1", "grp2-cmd1", "4"
            )[0].strip(),
            "test_app::adapted2(3)::grp2(True)::grp2_cmd1(4)",
        )

        self.assertEqual(
            run_command("adapted2", "--verbosity", "2", "grp2", "grp2-cmd2", "arg")[
                0
            ].strip(),
            "test_app::adapted2(2)::grp2(False)::grp2_cmd2(arg)",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "4", "grp1", "5", "grp1-ext"),
            "test_app::adapted2(4)::grp1(5.0)::grp1_ext()",
        )

        self.assertEqual(
            call_command("adapted2", "grp2", "grp2-cmd1", "arg"),
            "test_app::adapted2(1)::grp2(False)::grp2_cmd1(arg)",
        )

        with self.assertRaises(CommandError):
            # 6 is out of bounds
            call_command("adapted2", "--verbosity", "6", "grp1", "5", "grp1-ext")

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "a1", "a2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "top2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "grp1", "3", "grp1-adpt1")

        with self.assertRaises(CommandError):
            call_command("adapted2", "grp1", "4", "subgroup", "subsubgroup", "ssg-cmd")


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter0",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class AdapterTests(ResetAppsMixin, TestCase):
    def test_adapted_is_compound(self):
        self.assertTrue(get_command("adapted").is_compound_command)

    @pytest.mark.no_rich
    @pytest.mark.skipif(rich_installed, reason="rich is installed")
    def test_adapted_helps_no_rich(self):
        hlp = run_command("adapted", "--settings", "tests.settings.adapted", "--help")[
            0
        ]
        self.assertGreater(
            sim := similarity(adapted_help, hlp),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted --help similiarity: {sim}")

        hlp = run_command(
            "adapted",
            "--settings",
            "tests.settings.adapted",
            "adapted",
            "--help",
        )[0]
        self.assertGreater(
            sim := similarity(adapted_adapted_help, hlp),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted adapted --help similiarity: {sim}")

        hlp = run_command(
            "adapted",
            "--settings",
            "tests.settings.adapted",
            "echo",
            "--help",
        )[0]
        self.assertGreater(
            sim := similarity(adapted_echo_help, hlp),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted echo --help similiarity: {sim}")

    def test_adapted_runs(self):
        self.assertEqual(
            run_command(
                "adapted",
                "--settings",
                "tests.settings.adapted",
                "adapted",
                "hello",
            )[0].strip(),
            "test_app: hello",
        )
        self.assertEqual(
            run_command(
                "adapted",
                "--settings",
                "tests.settings.adapted",
                "echo",
                "hello",
                "world",
            )[0].strip(),
            "test_app2: hello world",
        )

    def test_adapted_calls(self):
        stdout = StringIO()
        call_command("adapted", "adapted", "hello", stdout=stdout)
        self.assertEqual(stdout.getvalue().strip(), "test_app: hello")
        stdout.seek(0)
        call_command("adapted", ["echo", "hello", "world"], stdout=stdout)
        self.assertEqual(stdout.getvalue().strip(), "test_app2: hello world")

    def test_adapted_direct(self):
        stdout = StringIO()
        adapted = get_command("adapted", stdout=stdout)
        adapted("hello")
        self.assertEqual(stdout.getvalue().strip(), "test_app: hello")
        stdout.seek(0)
        adapted.new_echo("hello", "world")
        self.assertEqual(stdout.getvalue().strip(), "test_app2: hello world")

    @pytest.mark.rich
    @pytest.mark.skipif(not rich_installed, reason="rich is not installed")
    def test_adapted_helps(self):
        stdout = StringIO()
        adapted = get_command("adapted", no_color=True, stdout=stdout)
        adapted.print_help("./manage.py", "adapted")
        self.assertGreater(
            sim := similarity(adapted_rich_help, stdout.getvalue()),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted --help similiarity: {sim}")

        stdout = StringIO()
        adapted = get_command("adapted", no_color=True, stdout=stdout)
        adapted.print_help("./manage.py", "adapted", "adapted")
        self.assertGreater(
            sim := similarity(adapted_adapted_rich_help, stdout.getvalue()),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted adapted --help similiarity: {sim}")

        stdout = StringIO()
        adapted = get_command("adapted", no_color=True, stdout=stdout)
        adapted.print_help("./manage.py", "adapted", "echo")
        self.assertGreater(
            sim := similarity(adapted_echo_rich_help, stdout.getvalue()),
            0.99,  # width inconsistences drive this number < 1
        )
        print(f"adapted echo --help similiarity: {sim}")


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter0",
        "tests.apps.test_app2",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class Adapter2AdaptedLevel1Tests(ResetAppsMixin, TestCase):
    def test_command_adapter_level1(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted",
                "--verbosity",
                "0",
                "top1",
                "arg1",
                "arg2",
            )[0].strip(),
            "test_app2::adapted2(0)::top1(arg1, arg2)",
        )
        self.assertEqual(
            call_command("adapted2", "--verbosity", "0", "top1", "arg1", "arg2"),
            "test_app2::adapted2(0)::top1(arg1, arg2)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "arg1")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        self.assertEqual(
            adapted2.top1("a1", "a2"), "test_app2::adapted2(3)::top1(a1, a2)"
        )

        with self.assertRaises(TypeError):
            adapted2.top1("a1")

    def test_group_adapter_level1(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted",
                "--verbosity",
                "0",
                "grp1",
                "3",
                "grp1-adpt1",
            )[0].strip(),
            "test_app2::adapted2(0)::grp1(3.0)::grp1_adpt1()",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp1", "2.5", "grp1-adpt1"),
            "test_app2::adapted2(2)::grp1(2.5)::grp1_adpt1()",
        )

        with self.assertRaises(CommandError):
            (call_command("adapted2", "grp1", "2.5", "grp1-adpt1", "5"),)

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp1(3.2)
        self.assertEqual(
            adapted2.grp1_adpt1(), "test_app2::adapted2(3)::grp1(3.2)::grp1_adpt1()"
        )

        with self.assertRaises(TypeError):
            adapted2.grp1_adpt1(4)

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted",
                "--verbosity",
                "0",
                "grp2",
                "--flag1",
                "sub-grp2",
            )[0].strip(),
            "test_app2::adapted2(0)::grp2(True)::sub_grp2()",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2"),
            "test_app2::adapted2(2)::grp2(False)::sub_grp2()",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "5")

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "5", "4")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp2(flag1=True)
        self.assertEqual(
            adapted2.sub_grp2(), "test_app2::adapted2(3)::grp2(True)::sub_grp2()"
        )

        with self.assertRaises(TypeError):
            adapted2.sub_grp2(4)

        self.assertFalse(hasattr(adapted2, "subsub_grp2"))
        with self.assertRaises(CommandError):
            call_command(
                "adapted2", "--verbosity", "2", "grp2", "sub-grp2", "subsub-grp2"
            )

        self.assertFalse(hasattr(adapted2, "grp3"))
        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp3")

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            )[0].strip(),
            "test_app::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
                arg5_0=True,
            ),
            "test_app::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::subsubcommand()",
        )

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
            )[0].strip(),
            "test_app2::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::ssg_cmd()",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                arg5_0=True,
            ),
            "test_app2::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::ssg_cmd()",
        )
        adapted2.grp1(3.4)
        adapted2.subgroup(True)
        self.assertEqual(
            adapted2.ssg_cmd(),
            "test_app2::interference::grp1(3.4)::subgroup(True)::subsubgroup(False)::ssg_cmd()",
        )


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter1",
        "tests.apps.adapter0",
        "tests.apps.test_app2",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class Adapter2AdaptedLevel2Tests(ResetAppsMixin, TestCase):
    def test_command_adapter_level2(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "top1",
                "arg1",
            )[0].strip(),
            "adapter1::adapted2(1)::top1(arg1)",
        )
        self.assertEqual(
            call_command("adapted2", "top1", "arg1"),
            "adapter1::adapted2(1)::top1(arg1)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "arg1", "arg2")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        self.assertEqual(adapted2.top1("a1"), "adapter1::adapted2(3)::top1(a1)")

        with self.assertRaises(TypeError):
            adapted2.top1()

    def test_group_adapter_level2(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "--verbosity",
                "0",
                "grp1",
                "3",
                "grp1-adpt1",
                "7",
            )[0].strip(),
            "adapter1::adapted2(0)::grp1(3.0)::grp1_adpt1(7)",
        )

        self.assertEqual(
            call_command(
                "adapted2", "--verbosity", "2", "grp1", "2.5", "grp1-adpt1", "12"
            ),
            "adapter1::adapted2(2)::grp1(2.5)::grp1_adpt1(12)",
        )

        with self.assertRaises(CommandError):
            (call_command("adapted2", "grp1", "2.5", "grp1-adpt1"),)

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp1(3.2)
        self.assertEqual(
            adapted2.grp1_adpt1(6), "adapter1::adapted2(3)::grp1(3.2)::grp1_adpt1(6)"
        )

        with self.assertRaises(TypeError):
            adapted2.grp1_adpt1()

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "--verbosity",
                "0",
                "grp2",
                "--flag1",
                "sub-grp2",
                "8",
            )[0].strip(),
            "adapter1::adapted2(0)::grp2(True)::sub_grp2(8)",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "9"),
            "adapter1::adapted2(2)::grp2(False)::sub_grp2(9)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "5", "4")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp2(flag1=True)
        self.assertEqual(
            adapted2.sub_grp2(3), "adapter1::adapted2(3)::grp2(True)::sub_grp2(3)"
        )

        with self.assertRaises(TypeError):
            adapted2.sub_grp2()

        self.assertFalse(hasattr(adapted2, "subsub_grp2"))
        with self.assertRaises(CommandError):
            call_command(
                "adapted2", "--verbosity", "2", "grp2", "sub-grp2", "subsub-grp2"
            )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp3"),
            "adapter1::adapted2(2)::grp3()",
        )
        self.assertEqual(adapted2.grp3(), "adapter1::adapted2(3)::grp3()")

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            )[0].strip(),
            "test_app::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
                arg5_0=True,
            ),
            "test_app::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::subsubcommand()",
        )

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "0",
            )[0].strip(),
            "adapter1::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::ssg_cmd(0)",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "2",
                arg5_0=True,
            ),
            "adapter1::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::ssg_cmd(2)",
        )
        adapted2.grp1(3.4)
        adapted2.subgroup(True)
        self.assertEqual(
            adapted2.ssg_cmd(5),
            "adapter1::interference::grp1(3.4)::subgroup(True)::subsubgroup(False)::ssg_cmd(5)",
        )


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter2",
        "tests.apps.adapter1",
        "tests.apps.adapter0",
        "tests.apps.test_app2",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class Adapter2AdaptedLevel3Tests(ResetAppsMixin, TestCase):
    def test_command_adapter_level3(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1_2",
                "top1",
            )[0].strip(),
            "adapter2::adapted2(1)::top1()",
        )
        self.assertEqual(
            call_command("adapted2", "top1", verbosity=4),
            "adapter2::adapted2(4)::top1()",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "arg1", "arg2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "arg1")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=2)
        self.assertEqual(adapted2.top1(), "adapter2::adapted2(2)::top1()")

        with self.assertRaises(TypeError):
            adapted2.top1("a1", "a2")

    def test_group_adapter_level3(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1_2",
                "--verbosity",
                "0",
                "grp1",
                "3",
                "grp1-adpt1",
                "7",
                "6",
            )[0].strip(),
            "adapter2::adapted2(0)::grp1(3.0)::grp1_adpt1(7, 6)",
        )

        self.assertEqual(
            call_command(
                "adapted2", "--verbosity", "2", "grp1", "2.5", "grp1-adpt1", "12", "8"
            ),
            "adapter2::adapted2(2)::grp1(2.5)::grp1_adpt1(12, 8)",
        )

        with self.assertRaises(CommandError):
            (call_command("adapted2", "grp1", "2.5", "grp1-adpt1"),)

        with self.assertRaises(CommandError):
            (call_command("adapted2", "grp1", "2.5", "grp1-adpt1", "12"),)

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp1(3.2)
        self.assertEqual(
            adapted2.grp1_adpt1(3, 6),
            "adapter2::adapted2(3)::grp1(3.2)::grp1_adpt1(3, 6)",
        )

        with self.assertRaises(TypeError):
            adapted2.grp1_adpt1()

        with self.assertRaises(TypeError):
            adapted2.grp1_adpt1(55)

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1_2",
                "--verbosity",
                "0",
                "grp2",
                "--flag1",
                "sub-grp2",
                "8",
                "7",
            )[0].strip(),
            "adapter2::adapted2(0)::grp2(True)::sub_grp2(8, 7)",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "9", "7"),
            "adapter2::adapted2(2)::grp2(False)::sub_grp2(9, 7)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "5")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp2(flag1=True)
        self.assertEqual(
            adapted2.sub_grp2(6, 3),
            "adapter2::adapted2(3)::grp2(True)::sub_grp2(6, 3)",
        )

        with self.assertRaises(TypeError):
            adapted2.sub_grp2()

        with self.assertRaises(TypeError):
            adapted2.sub_grp2(3)

        self.assertEqual(
            call_command(
                "adapted2",
                "--verbosity",
                "2",
                "grp2",
                "sub-grp2",
                "4",
                "5",
                "subsub-grp2",
            ),
            "adapter2::adapted2()::grp2()::sub_grp2()::subsub_grp2()",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp3"),
            "adapter1::adapted2(2)::grp3()",
        )
        self.assertEqual(adapted2.grp3(), "adapter1::adapted2(3)::grp3()")

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1_2",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            )[0].strip(),
            "test_app::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
                arg5_0=True,
            ),
            "test_app::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::subsubcommand()",
        )

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1_2",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "0",
                "1",
            )[0].strip(),
            "adapter2::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::ssg_cmd(0, 1)",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "2",
                "3",
                arg5_0=True,
            ),
            "adapter2::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::ssg_cmd(2, 3)",
        )
        adapted2.grp1(3.4)
        adapted2.subgroup(True)
        self.assertEqual(
            adapted2.ssg_cmd(5, 6),
            "adapter2::interference::grp1(3.4)::subgroup(True)::subsubgroup(False)::ssg_cmd(5, 6)",
        )


@override_settings(
    INSTALLED_APPS=[
        "tests.apps.adapter1",
        "tests.apps.adapter2",
        "tests.apps.adapter0",
        "tests.apps.test_app2",
        "tests.apps.test_app",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
    ]
)
class Adapter2AdaptedPrecedenceTests(ResetAppsMixin, TestCase):
    """
    Reverse app stack order and test that we observe the correct app
    winning out.
    """

    def test_command_adapter_precedence(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted2_1",
                "top1",
                "arg1",
            )[0].strip(),
            "adapter1::adapted2(1)::top1(arg1)",
        )
        self.assertEqual(
            call_command("adapted2", "top1", "arg1"),
            "adapter1::adapted2(1)::top1(arg1)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "top1", "arg1", "arg2")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        self.assertEqual(adapted2.top1("a1"), "adapter1::adapted2(3)::top1(a1)")

        with self.assertRaises(TypeError):
            adapted2.top1()

    def test_group_adapter_precedence(self):
        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "--verbosity",
                "0",
                "grp1",
                "3",
                "grp1-adpt1",
                "7",
            )[0].strip(),
            "adapter1::adapted2(0)::grp1(3.0)::grp1_adpt1(7)",
        )

        self.assertEqual(
            call_command(
                "adapted2", "--verbosity", "2", "grp1", "2.5", "grp1-adpt1", "12"
            ),
            "adapter1::adapted2(2)::grp1(2.5)::grp1_adpt1(12)",
        )

        with self.assertRaises(CommandError):
            (call_command("adapted2", "grp1", "2.5", "grp1-adpt1"),)

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp1(3.2)
        self.assertEqual(
            adapted2.grp1_adpt1(6), "adapter1::adapted2(3)::grp1(3.2)::grp1_adpt1(6)"
        )

        with self.assertRaises(TypeError):
            adapted2.grp1_adpt1()

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "--verbosity",
                "0",
                "grp2",
                "--flag1",
                "sub-grp2",
                "8",
            )[0].strip(),
            "adapter1::adapted2(0)::grp2(True)::sub_grp2(8)",
        )

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "9"),
            "adapter1::adapted2(2)::grp2(False)::sub_grp2(9)",
        )

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2")

        with self.assertRaises(CommandError):
            call_command("adapted2", "--verbosity", "2", "grp2", "sub-grp2", "5", "4")

        adapted2 = get_command("adapted2")
        adapted2.init(verbosity=3)
        adapted2.grp2(flag1=True)
        self.assertEqual(
            adapted2.sub_grp2(3), "adapter1::adapted2(3)::grp2(True)::sub_grp2(3)"
        )

        with self.assertRaises(TypeError):
            adapted2.sub_grp2()

        # this doesnt work because adapter1 overrides the whole subgroup
        with self.assertRaises(CommandError):
            call_command(
                "adapted2", "--verbosity", "2", "grp2", "sub-grp2", "4", "subsub-grp2"
            )

        self.assertFalse(hasattr(adapted2.__class__, "subsub_grp2"))
        self.assertFalse(hasattr(adapted2, "subsub_grp2"))

        self.assertEqual(
            call_command("adapted2", "--verbosity", "2", "grp3"),
            "adapter1::adapted2(2)::grp3()",
        )
        self.assertEqual(adapted2.grp3(), "adapter1::adapted2(3)::grp3()")

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
            )[0].strip(),
            "test_app::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::subsubcommand()",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "subsubcommand",
                arg5_0=True,
            ),
            "test_app::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::subsubcommand()",
        )

        self.assertEqual(
            run_command(
                "adapted2",
                "--settings",
                "tests.settings.adapted1",
                "grp1",
                "5",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "0",
            )[0].strip(),
            "adapter1::interference::grp1(5.0)::subgroup(False)::subsubgroup(True)::ssg_cmd(0)",
        )
        self.assertEqual(
            call_command(
                "adapted2",
                "grp1",
                "10",
                "subgroup",
                "subsubgroup",
                "ssg-cmd",
                "2",
                arg5_0=True,
            ),
            "adapter1::interference::grp1(10.0)::subgroup(True)::subsubgroup(True)::ssg_cmd(2)",
        )
        adapted2.grp1(3.4)
        adapted2.subgroup(True)
        self.assertEqual(
            adapted2.ssg_cmd(5),
            "adapter1::interference::grp1(3.4)::subgroup(True)::subsubgroup(False)::ssg_cmd(5)",
        )



import json
import sys
from typing import Annotated

from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import TyperCommand, group


class Command(TyperCommand):
    help = _("Test overloaded option and argument names")

    parameters = {}

    @group()
    def test(
        self, precision: int, flag: Annotated[bool, Option(help="A flag.")] = False
    ):
        """
        Do some math at the given precision.
        """
        assert self.__class__ is Command
        self.parameters = {}
        self.parameters["test"] = {
            "precision": precision,
            "flag": flag,
        }

    @test.command()
    def samename(
        self, precision: int, flag: Annotated[bool, Option(help="A flag.")] = False
    ):
        assert self.__class__ is Command
        self.parameters["samename"] = {
            "precision": precision,
            "flag": flag,
        }
        return json.dumps(self.parameters, indent=2)

    @test.command()
    def diffname(
        self,
        precision2: Annotated[int, Argument(metavar="precision")],
        flag2: Annotated[bool, Option("--flag/--no-flag", help="A flag.")] = False,
    ):
        assert self.__class__ is Command
        self.parameters["diffname"] = {
            "precision2": precision2,
            "flag2": flag2,
        }
        return json.dumps(self.parameters, indent=2)
